# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from osc_lib.command import command
from osc_lib import utils


class ListService(command.Lister):
    """Show a list of all running services."""

    log = logging.getLogger(__name__ + ".ListService")

    def get_parser(self, prog_name):
        parser = super(ListService, self).get_parser(prog_name)
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        senlin_client = self.app.client_manager.clustering
        columns = ['binary', 'host', 'status', 'state', 'updated_at',
                   'disabled_reason']

        queries = {}
        result = senlin_client.services(**queries)

        formatters = {}
        return (
            columns,
            (utils.get_item_properties(s, columns, formatters=formatters)
             for s in result)
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from openstack import exceptions as sdk_exc
from osc_lib import exceptions as exc
from unittest import mock

from senlinclient.tests.unit.v1 import fakes
from senlinclient.v1 import policy_type as osc_policy_type


class TestPolicyType(fakes.TestClusteringv1):
    def setUp(self):
        super(TestPolicyType, self).setUp()
        self.mock_client = self.app.client_manager.clustering


class TestPolicyTypeList(TestPolicyType):
    expected_columns = ['name', 'version', 'support_status']
    pt1 = mock.Mock(
        schema={'foo': 'bar'},
        support_status={
            "1.0": [{"status": "SUPPORTED", "since": "2016.10"}]
        }
    )
    pt1.name = 'BBB'
    pt2 = mock.Mock(
        schema={'foo': 'bar'},
        support_status={
            "1.0": [{"status": "DEPRECATED", "since": "2016.01"}]
        }
    )
    pt2.name = 'AAA'
    list_response = [pt1, pt2]
    expected_rows = [
        ('AAA', '1.0', 'DEPRECATED since 2016.01'),
        ('BBB', '1.0', 'SUPPORTED since 2016.10')
    ]

    def setUp(self):
        super(TestPolicyTypeList, self).setUp()
        self.cmd = osc_policy_type.PolicyTypeList(self.app, None)
        self.mock_client.policy_types = mock.Mock(
            return_value=self.list_response)

    def test_policy_type_list(self):
        arglist = []
        parsed_args = self.check_parser(self.cmd, arglist, [])
        expected_columns = self.expected_columns
        expected_rows = self.expected_rows
        columns, rows = self.cmd.take_action(parsed_args)
        if len(columns) == 2:
            expected_columns = ['name', 'version']
            expected_rows = [
                ('CCC', '1.0')
            ]

        self.mock_client.policy_types.assert_called_with()
        self.assertEqual(expected_columns, columns)
        self.assertEqual(expected_rows, rows)


class TestPolicyTypeShow(TestPolicyType):

    def setUp(self):
        super(TestPolicyTypeShow, self).setUp()
        self.cmd = osc_policy_type.PolicyTypeShow(self.app, None)
        fake_pt = mock.Mock(schema={'foo': 'bar'})
        fake_pt.name = 'senlin.policy.deletion-1.0'
        self.mock_client.get_policy_type = mock.Mock(return_value=fake_pt)

    def test_policy_type_show(self):
        arglist = ['os.heat.stack-1.0']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.get_policy_type.assert_called_once_with(
            'os.heat.stack-1.0')

    def test_policy_type_show_not_found(self):
        arglist = ['senlin.policy.deletion-1.0']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.mock_client.get_policy_type.side_effect = (
            sdk_exc.ResourceNotFound())
        error = self.assertRaises(exc.CommandError, self.cmd.take_action,
                                  parsed_args)
        self.assertEqual('Policy Type not found: senlin.policy.deletion-1.0',
                         str(error))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""OpenStackClient plugin for Clustering service."""

import logging

from openstack.config import cloud_region
from openstack.config import defaults as config_defaults
from openstack import connection
from osc_lib import utils

LOG = logging.getLogger(__name__)

DEFAULT_CLUSTERING_API_VERSION = '1'
API_VERSION_OPTION = 'os_clustering_api_version'
API_NAME = 'clustering'
CURRENT_API_VERSION = '1.14'


def _make_key(service_type, key):
    if not service_type:
        return key
    else:
        service_type = service_type.lower().replace('-', '_')
        return "_".join([service_type, key])


def _get_config_from_profile(profile, **kwargs):
    # Deal with clients still trying to use legacy profile objects
    region_name = None
    for service in profile.get_services():
        if service.region:
            region_name = service.region
        service_type = service.service_type
        if service.interface:
            key = _make_key(service_type, 'interface')
            kwargs[key] = service.interface
        if service.version:
            version = service.version
            if version.startswith('v'):
                version = version[1:]
            key = _make_key(service_type, 'api_version')
            kwargs[key] = version
        if service.api_version:
            version = service.api_version
            key = _make_key(service_type, 'default_microversion')
            kwargs[key] = version

    config_kwargs = config_defaults.get_defaults()
    config_kwargs.update(kwargs)
    config = cloud_region.CloudRegion(
        region_name=region_name, config=config_kwargs)
    return config


def create_connection(prof=None, cloud_region=None, **kwargs):
    version_key = _make_key(API_NAME, 'api_version')
    kwargs[version_key] = CURRENT_API_VERSION

    if not cloud_region:
        if prof:
            cloud_region = _get_config_from_profile(prof, **kwargs)
    else:
        # If we got the CloudRegion from python-openstackclient and it doesn't
        # already have a default microversion set, set it here.
        microversion_key = _make_key(API_NAME, 'default_microversion')
        cloud_region.config.setdefault(microversion_key, CURRENT_API_VERSION)

    user_agent = kwargs.pop('user_agent', None)
    app_name = kwargs.pop('app_name', None)
    app_version = kwargs.pop('app_version', None)
    if user_agent is not None and (not app_name and not app_version):
        app_name, app_version = user_agent.split('/', 1)

    return connection.Connection(
        config=cloud_region,
        app_name=app_name,
        app_version=app_version, **kwargs)


def make_client(instance):
    """Returns a clustering proxy"""
    # TODO(mordred) the ClientManager already has an OpenStackSDK connection,
    # but it only has it once setup_auth has been called. For things that
    # don't require auth, this is problematic, so we have to make our own.
    # Use the CloudRegion stored on the ClientManager for now.
    conn = create_connection(
        cloud_region=instance._cli_options,
    )

    LOG.debug('Connection: %s', conn)
    LOG.debug('Clustering client initialized using OpenStackSDK: %s',
              conn.clustering)
    return conn.clustering


def build_option_parser(parser):
    """Hook to add global options"""
    parser.add_argument(
        '--os-clustering-api-version',
        metavar='<clustering-api-version>',
        default=utils.env(
            'OS_CLUSTERING_API_VERSION',
            default=DEFAULT_CLUSTERING_API_VERSION),
        help='Clustering API version, default=' +
             DEFAULT_CLUSTERING_API_VERSION +
             ' (Env: OS_CLUSTERING_API_VERSION)')
    return parser



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Clustering v1 cluster action implementations"""

import logging
import subprocess
import sys
import threading
import time

from openstack import exceptions as sdk_exc
from osc_lib.command import command
from osc_lib import exceptions as exc
from osc_lib import utils
from oslo_utils import strutils

from senlinclient.common.i18n import _
from senlinclient.common import utils as senlin_utils


class ListCluster(command.Lister):
    """List clusters."""

    log = logging.getLogger(__name__ + ".ListCluster")

    def get_parser(self, prog_name):
        parser = super(ListCluster, self).get_parser(prog_name)
        parser.add_argument(
            '--filters',
            metavar='<"key1=value1;key2=value2...">',
            help=_("Filter parameters to apply on returned clusters. "
                   "This can be specified multiple times, or once with "
                   "parameters separated by a semicolon. The valid filter"
                   " keys are: ['status', 'name']"),
            action='append'
        )
        parser.add_argument(
            '--sort',
            metavar='<key>[:<direction>]',
            help=_("Sorting option which is a string containing a list of "
                   "keys separated by commas. Each key can be optionally "
                   "appended by a sort direction (:asc or :desc). The valid "
                   "sort keys are: ['name', 'status', 'init_at', "
                   "'created_at', 'updated_at']"))
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            help=_('Limit the number of clusters returned')
        )
        parser.add_argument(
            '--marker',
            metavar='<id>',
            help=_('Only return clusters that appear after the given cluster '
                   'ID')
        )
        parser.add_argument(
            '--global-project',
            default=False,
            action="store_true",
            help=_('Indicate that the cluster list should include clusters '
                   'from all projects. This option is subject to access '
                   'policy checking. Default is False')
        )
        parser.add_argument(
            '--full-id',
            default=False,
            action="store_true",
            help=_('Print full IDs in list')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        senlin_client = self.app.client_manager.clustering
        columns = ['id', 'name', 'status', 'created_at', 'updated_at']
        queries = {
            'limit': parsed_args.limit,
            'marker': parsed_args.marker,
            'sort': parsed_args.sort,
            'global_project': parsed_args.global_project,
        }
        if parsed_args.filters:
            queries.update(senlin_utils.format_parameters(parsed_args.filters))

        clusters = senlin_client.clusters(**queries)
        formatters = {}
        if parsed_args.global_project:
            columns.append('project_id')
        if not parsed_args.full_id:
            formatters = {
                'id': lambda x: x[:8]
            }
            if 'project_id' in columns:
                formatters['project_id'] = lambda x: x[:8]

        return (
            columns,
            (utils.get_item_properties(c, columns, formatters=formatters)
             for c in clusters)
        )


class ShowCluster(command.ShowOne):
    """Show details of the cluster."""

    log = logging.getLogger(__name__ + ".ShowCluster")

    def get_parser(self, prog_name):
        parser = super(ShowCluster, self).get_parser(prog_name)
        parser.add_argument(
            'cluster',
            metavar='<cluster>',
            help=_('Name or ID of cluster to show')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        senlin_client = self.app.client_manager.clustering
        return _show_cluster(senlin_client, parsed_args.cluster)


def _show_cluster(senlin_client, cluster_id):
    try:
        cluster = senlin_client.get_cluster(cluster_id)
    except sdk_exc.ResourceNotFound:
        raise exc.CommandError(_('Cluster not found: %s') % cluster_id)

    formatters = {
        'config': senlin_utils.json_formatter,
        'metadata': senlin_utils.json_formatter,
        'node_ids': senlin_utils.list_formatter
    }
    data = cluster.to_dict()
    if 'is_profile_only' in data:
        data.pop('is_profile_only')
    columns = sorted(data.keys())
    return columns, utils.get_dict_properties(data, columns,
                                              formatters=formatters)


class CreateCluster(command.ShowOne):
    """Create the cluster."""

    log = logging.getLogger(__name__ + ".CreateCluster")

    def get_parser(self, prog_name):
        parser = super(CreateCluster, self).get_parser(prog_name)
        parser.add_argument(
            '--config',
            metavar='<"key1=value1;key2=value2...">',
            help=_('Configuration of the cluster. Default to {}. '
                   'This can be specified multiple times, or once with '
                   'key-value pairs separated by a semicolon.'),
            action='append'
        )
        parser.add_argument(
            '--min-size',
            metavar='<min-size>',
            default=0,
            help=_('Min size of the cluster. Default to 0')
        )
        parser.add_argument(
            '--max-size',
            metavar='<max-size>',
            default=-1,
            help=_('Max size of the cluster. Default to -1, means unlimited')
        )
        parser.add_argument(
            '--desired-capacity',
            metavar='<desired-capacity>',
            default=0,
            help=_('Desired capacity of the cluster. Default to min_size if '
                   'min_size is specified else 0.')
        )
        parser.add_argument(
            '--timeout',
            metavar='<timeout>',
            type=int,
            help=_('Cluster creation timeout in seconds')
        )
        parser.add_argument(
            '--metadata',
            metavar='<"key1=value1;key2=value2...">',
            help=_('Metadata values to be attached to the cluster. '
                   'This can be specified multiple times, or once with '
                   'key-value pairs separated by a semicolon.'),
            action='append'
        )
        parser.add_argument(
            '--profile',
            metavar='<profile>',
            required=True,
            help=_('Default profile Id or name used for this cluster')
        )
        parser.add_argument(
            'name',
            metavar='<cluster-name>',
            help=_('Name of the cluster to create')
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help=_('Wait for cluster creation to complete')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        senlin_client = self.app.client_manager.clustering
        if parsed_args.min_size and not parsed_args.desired_capacity:
            parsed_args.desired_capacity = parsed_args.min_size
        attrs = {
            'config': senlin_utils.format_parameters(parsed_args.config),
            'name': parsed_args.name,
            'profile_id': parsed_args.profile,
            'min_size': parsed_args.min_size,
            'max_size': parsed_args.max_size,
            'desired_capacity': parsed_args.desired_capacity,
            'metadata': senlin_utils.format_parameters(parsed_args.metadata),
            'timeout': parsed_args.timeout
        }

        cluster = senlin_client.create_cluster(**attrs)
        if parsed_args.wait:
            senlin_utils.await_cluster_status(senlin_client, cluster.id)
        return _show_cluster(senlin_client, cluster.id)


class UpdateCluster(command.ShowOne):
    """Update the cluster."""

    log = logging.getLogger(__name__ + ".UpdateCluster")

    def get_parser(self, prog_name):
        parser = super(UpdateCluster, self).get_parser(prog_name)
        parser.add_argument(
            '--config',
            metavar='<"key1=value1;key2=value2...">',
            help=_('Configuration of the cluster. '
                   'This can be specified multiple times, or once with '
                   'key-value pairs separated by a semicolon. '
                   'Any existing configuration values on the cluster are '
                   'wiped out when using this option.'),
            action='append'
        )
        parser.add_argument(
            '--profile',
            metavar='<profile>',
            help=_('ID or name of new profile to use')
        )
        parser.add_argument(
            '--profile-only',
            default=False, metavar='<boolean>',
            help=_("Whether the cluster should be updated profile only. "
                   "If false, it will be applied to all existing nodes. "
                   "If true, any newly created nodes will use the new profile,"
                   "but existing nodes will not be changed. Default is False.")
        )
        parser.add_argument(
            '--timeout',
            metavar='<timeout>',
            help=_('New timeout (in seconds) value for the cluster')
        )
        parser.add_argument(
            '--metadata',
            metavar='<"key1=value1;key2=value2...">',
            help=_("Metadata values to be attached to the cluster. "
                   "This can be specified multiple times, or once with "
                   "key-value pairs separated by a semicolon. Use '{}' "
                   "can clean metadata "),
            action='append'
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('New name for the cluster to update')
        )
        parser.add_argument(
            'cluster',
            metavar='<cluster>',
            help=_('Name or ID of cluster to be updated')
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help=_('Wait for cluster update to complete')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering
        cluster = senlin_client.find_cluster(parsed_args.cluster)
        if cluster is None:
            raise exc.CommandError(_('Cluster not found: %s') %
                                   parsed_args.cluster)
        attrs = {
            'name': parsed_args.name,
            'profile_id': parsed_args.profile,
            'profile_only': strutils.bool_from_string(
                parsed_args.profile_only,
                strict=True,
            ),
            'timeout': parsed_args.timeout,
        }

        if parsed_args.config is not None:
            attrs['config'] = senlin_utils.format_parameters(
                parsed_args.config)
        if parsed_args.metadata is not None:
            attrs['metadata'] = senlin_utils.format_parameters(
                parsed_args.metadata)

        senlin_client.update_cluster(cluster, **attrs)
        if parsed_args.wait:
            # PATCH operations do not currently return an action to await.
            # introducing a delay to allow the cluster to transition state
            # out of ACTIVE before inspection.
            time.sleep(1)
            senlin_utils.await_cluster_status(senlin_client, cluster.id)
        return _show_cluster(senlin_client, cluster.id)


class DeleteCluster(command.Command):
    """Delete the cluster(s)."""

    log = logging.getLogger(__name__ + ".DeleteCluster")

    def get_parser(self, prog_name):
        parser = super(DeleteCluster, self).get_parser(prog_name)
        parser.add_argument(
            'cluster',
            metavar='<cluster>',
            nargs='+',
            help=_('Name or ID of cluster(s) to delete.')
        )
        parser.add_argument(
            '--force-delete',
            action='store_true',
            help=_('Force to delete cluster(s).')
        )
        parser.add_argument(
            '--force',
            action='store_true',
            help=_('Skip yes/no prompt (assume yes).')
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help=_('Wait for cluster delete to complete')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering

        try:
            if not parsed_args.force and sys.stdin.isatty():
                sys.stdout.write(
                    _("Are you sure you want to delete this cluster(s)"
                      " [y/N]?"))
                sys.stdout.flush()
                prompt_response = sys.stdin.readline().lower()
                if not prompt_response.startswith('y'):
                    return
        except KeyboardInterrupt:  # Ctrl-c
            self.log.info('Ctrl-c detected.')
            return
        except EOFError:  # Ctrl-d
            self.log.info('Ctrl-d detected')
            return

        result = {}
        for cid in parsed_args.cluster:
            try:
                action = senlin_client.delete_cluster(
                    cid, False, parsed_args.force_delete)
                result[cid] = ('OK', action['id'])
            except Exception as ex:
                result[cid] = ('ERROR', str(ex))

        for cid, a in result.items():
            senlin_utils.print_action_result(cid, a)
            if parsed_args.wait:
                if a[0] == 'OK':
                    senlin_utils.await_action(senlin_client, a[1])
                    senlin_utils.await_cluster_delete(senlin_client, cid)


class ResizeCluster(command.ShowOne):
    """Resize a cluster."""

    log = logging.getLogger(__name__ + ".ResizeCluster")

    def get_parser(self, prog_name):
        parser = super(ResizeCluster, self).get_parser(prog_name)
        parser.add_argument(
            '--capacity',
            metavar='<capacity>',
            type=int,
            help=_('The desired number of nodes of the cluster')
        )
        parser.add_argument(
            '--adjustment',
            metavar='<adjustment>',
            type=int,
            help=_('A positive integer meaning the number of nodes to add, '
                   'or a negative integer indicating the number of nodes to '
                   'remove')
        )
        parser.add_argument(
            '--percentage',
            metavar='<percentage>',
            type=float,
            help=_('A value that is interpreted as the percentage of size '
                   'adjustment. This value can be positive or negative')
        )
        parser.add_argument(
            '--min-step',
            metavar='<min-step>',
            type=int,
            help=_('An integer specifying the number of nodes for adjustment '
                   'when <PERCENTAGE> is specified')
        )
        parser.add_argument(
            '--strict',
            action='store_true',
            default=False,
            help=_('A boolean specifying whether the resize should be '
                   'performed on a best-effort basis when the new capacity '
                   'may go beyond size constraints')
        )
        parser.add_argument(
            '--min-size',
            metavar='min',
            type=int,
            help=_('New lower bound of cluster size')
        )
        parser.add_argument(
            '--max-size',
            metavar='max',
            type=int,
            help=_('New upper bound of cluster size. A value of -1 indicates '
                   'no upper limit on cluster size')
        )
        parser.add_argument(
            'cluster',
            metavar='<cluster>',
            help=_('Name or ID of cluster to operate on')
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help=_('Wait for cluster resize to complete')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering

        action_args = {}

        capacity = parsed_args.capacity
        adjustment = parsed_args.adjustment
        percentage = parsed_args.percentage
        min_size = parsed_args.min_size
        max_size = parsed_args.max_size
        min_step = parsed_args.min_step
        wait = parsed_args.wait

        if sum(v is not None for v in (capacity, adjustment, percentage,
                                       min_size, max_size)) == 0:
            raise exc.CommandError(_("At least one parameter of 'capacity', "
                                     "'adjustment', 'percentage', 'min_size' "
                                     "and 'max_size' should be specified."))

        if sum(v is not None for v in (capacity, adjustment, percentage)) > 1:
            raise exc.CommandError(_("Only one of 'capacity', 'adjustment' and"
                                     " 'percentage' can be specified."))

        action_args['adjustment_type'] = None
        action_args['number'] = None

        if capacity is not None:
            if capacity < 0:
                raise exc.CommandError(_('Cluster capacity must be larger than'
                                         ' or equal to zero.'))
            action_args['adjustment_type'] = 'EXACT_CAPACITY'
            action_args['number'] = capacity

        if adjustment is not None:
            if adjustment == 0:
                raise exc.CommandError(_('Adjustment cannot be zero.'))
            action_args['adjustment_type'] = 'CHANGE_IN_CAPACITY'
            action_args['number'] = adjustment

        if percentage is not None:
            if (percentage == 0 or percentage == 0.0):
                raise exc.CommandError(_('Percentage cannot be zero.'))
            action_args['adjustment_type'] = 'CHANGE_IN_PERCENTAGE'
            action_args['number'] = percentage

        if min_step is not None and percentage is None:
            raise exc.CommandError(_('Min step is only used with '
                                     'percentage.'))

        if min_size is not None:
            if min_size < 0:
                raise exc.CommandError(_('Min size cannot be less than zero.'))
            if max_size is not None and max_size >= 0 and min_size > max_size:
                raise exc.CommandError(_('Min size cannot be larger than '
                                         'max size.'))
            if capacity is not None and min_size > capacity:
                raise exc.CommandError(_('Min size cannot be larger than the '
                                         'specified capacity'))

        if max_size is not None:
            if capacity is not None and max_size > 0 and max_size < capacity:
                raise exc.CommandError(_('Max size cannot be less than the '
                                         'specified capacity.'))
            # do a normalization
            if max_size < 0:
                max_size = -1

        action_args['min_size'] = min_size
        action_args['max_size'] = max_size
        action_args['min_step'] = min_step
        action_args['strict'] = parsed_args.strict

        resp = senlin_client.resize_cluster(parsed_args.cluster, **action_args)

        if 'action' in resp:
            print('Request accepted by action: %s' % resp['action'])
            if wait:
                senlin_utils.await_action(senlin_client, resp['action'])
                senlin_utils.await_cluster_status(senlin_client,
                                                  parsed_args.cluster)
                return _show_cluster(senlin_client, parsed_args.cluster)
        else:
            print('Request error: %s' % resp)

        return '', ''


class ScaleInCluster(command.ShowOne):
    """Scale in a cluster by the specified number of nodes."""

    log = logging.getLogger(__name__ + ".ScaleInCluster")

    def get_parser(self, prog_name):
        parser = super(ScaleInCluster, self).get_parser(prog_name)
        parser.add_argument(
            '--count',
            metavar='<count>',
            help=_('Number of nodes to be deleted from the specified cluster')
        )
        parser.add_argument(
            'cluster',
            metavar='<cluster>',
            help=_('Name or ID of cluster to operate on')
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help=_('Wait for cluster scale-in to complete')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering

        resp = senlin_client.scale_in_cluster(parsed_args.cluster,
                                              parsed_args.count)
        status_code = resp.get('code', None)
        if status_code in [409]:
            raise exc.CommandError(_(
                'Unable to scale in cluster: %s') % resp['error']['message'])
        if 'action' in resp:
            print('Request accepted by action: %s' % resp['action'])
            if parsed_args.wait:
                senlin_utils.await_action(senlin_client, resp['action'])
                senlin_utils.await_cluster_status(senlin_client,
                                                  parsed_args.cluster)
                return _show_cluster(senlin_client, parsed_args.cluster)
        else:
            print('Request error: %s' % resp)

        return '', ''


class ScaleOutCluster(command.ShowOne):
    """Scale out a cluster by the specified number of nodes."""

    log = logging.getLogger(__name__ + ".ScaleOutCluster")

    def get_parser(self, prog_name):
        parser = super(ScaleOutCluster, self).get_parser(prog_name)
        parser.add_argument(
            '--count',
            metavar='<count>',
            help=_('Number of nodes to be added to the specified cluster')
        )
        parser.add_argument(
            'cluster',
            metavar='<cluster>',
            help=_('Name or ID of cluster to operate on')
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help=_('Wait for cluster scale-out to complete')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering

        resp = senlin_client.scale_out_cluster(parsed_args.cluster,
                                               parsed_args.count)
        status_code = resp.get('code', None)
        if status_code in [409]:
            raise exc.CommandError(_(
                'Unable to scale out cluster: %s') % resp['error']['message'])
        if 'action' in resp:
            print('Request accepted by action: %s' % resp['action'])
            if parsed_args.wait:
                senlin_utils.await_action(senlin_client, resp['action'])
                senlin_utils.await_cluster_status(senlin_client,
                                                  parsed_args.cluster)
                return _show_cluster(senlin_client, parsed_args.cluster)
        else:
            print('Request error: %s' % resp)

        return '', ''


class ClusterPolicyAttach(command.Command):
    """Attach policy to cluster."""

    log = logging.getLogger(__name__ + ".ClusterPolicyAttach")

    def get_parser(self, prog_name):
        parser = super(ClusterPolicyAttach, self).get_parser(prog_name)
        parser.add_argument(
            '--enabled',
            metavar='<boolean>',
            default=True,
            help=_('Whether the policy should be enabled once attached. '
                   'Default to True')
        )
        parser.add_argument(
            '--policy',
            metavar='<policy>',
            required=True,
            help=_('ID or name of policy to be attached')
        )
        parser.add_argument(
            'cluster',
            metavar='<cluster>',
            help=_('Name or ID of cluster to operate on')
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help=_('Wait for cluster policy-attach to complete')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering

        kwargs = {
            'enabled': strutils.bool_from_string(parsed_args.enabled,
                                                 strict=True),
        }

        resp = senlin_client.attach_policy_to_cluster(parsed_args.cluster,
                                                      parsed_args.policy,
                                                      **kwargs)
        if 'action' in resp:
            print('Request accepted by action: %s' % resp['action'])
            if parsed_args.wait:
                senlin_utils.await_action(senlin_client, resp['action'])
        else:
            print('Request error: %s' % resp)


class ClusterPolicyDetach(command.Command):
    """Detach policy from cluster."""

    log = logging.getLogger(__name__ + ".ClusterPolicyDetach")

    def get_parser(self, prog_name):
        parser = super(ClusterPolicyDetach, self).get_parser(prog_name)
        parser.add_argument(
            '--policy',
            metavar='<policy>',
            required=True,
            help=_('ID or name of policy to be detached')
        )
        parser.add_argument(
            'cluster',
            metavar='<cluster>',
            help=_('Name or ID of cluster to operate on')
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help=_('Wait for cluster policy-detach to complete')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering
        resp = senlin_client.detach_policy_from_cluster(parsed_args.cluster,
                                                        parsed_args.policy)
        if 'action' in resp:
            print('Request accepted by action: %s' % resp['action'])
            if parsed_args.wait:
                senlin_utils.await_action(senlin_client, resp['action'])
        else:
            print('Request error: %s' % resp)


class ClusterNodeList(command.Lister):
    """List nodes from cluster."""

    log = logging.getLogger(__name__ + ".ClusterNodeList")

    def get_parser(self, prog_name):
        parser = super(ClusterNodeList, self).get_parser(prog_name)
        parser.add_argument(
            '--filters',
            metavar='<key1=value1;key2=value2...>',
            help=_("Filter parameters to apply on returned nodes. "
                   "This can be specified multiple times, or once with "
                   "parameters separated by a semicolon. The valid filter "
                   "keys are: ['status', 'name']"),
            action='append'
        )
        parser.add_argument(
            '--sort',
            metavar='<key>[:<direction>]',
            help=_("Sorting option which is a string containing a list of "
                   "keys separated by commas. Each key can be optionally "
                   "appended by a sort direction (:asc or :desc)' The valid "
                   "sort keys are:['index', 'name', 'status', 'init_at', "
                   "'created_at', 'updated_at']")
        )
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            help=_('Limit the number of nodes returned')
        )
        parser.add_argument(
            '--marker',
            metavar='<id>',
            help=_('Only return nodes that appear after the given node ID')
        )
        parser.add_argument(
            '--full-id',
            default=False,
            action="store_true",
            help=_('Print full IDs in list')
        )
        parser.add_argument(
            'cluster',
            metavar='<cluster>',
            help=_('Name or ID of cluster to nodes from')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering
        queries = {
            'cluster_id': parsed_args.cluster,
            'sort': parsed_args.sort,
            'limit': parsed_args.limit,
            'marker': parsed_args.marker,
        }
        if parsed_args.filters:
            queries.update(senlin_utils.format_parameters(parsed_args.filters))

        nodes = senlin_client.nodes(**queries)
        if not parsed_args.full_id:
            formatters = {
                'id': lambda x: x[:8],
                'physical_id': lambda x: x[:8] if x else ''
            }
        else:
            formatters = {}

        columns = ['id', 'name', 'index', 'status', 'physical_id',
                   'created_at']
        return (
            columns,
            (utils.get_item_properties(n, columns, formatters=formatters)
             for n in nodes)
        )


class ClusterNodeAdd(command.ShowOne):
    """Add specified nodes to cluster."""
    log = logging.getLogger(__name__ + ".ClusterNodeAdd")

    def get_parser(self, prog_name):
        parser = super(ClusterNodeAdd, self).get_parser(prog_name)
        parser.add_argument(
            '--nodes',
            metavar='<nodes>',
            required=True,
            help=_('ID or name of nodes to be added; multiple nodes can be'
                   ' separated with ","')
        )
        parser.add_argument(
            'cluster',
            metavar='<cluster>',
            help=_('Name or ID of cluster to operate on')
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help=_('Wait for cluster members add to complete')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering
        node_ids = parsed_args.nodes.split(',')
        resp = senlin_client.add_nodes_to_cluster(parsed_args.cluster,
                                                  node_ids)
        if 'action' in resp:
            print('Request accepted by action: %s' % resp['action'])
            if parsed_args.wait:
                senlin_utils.await_action(senlin_client, resp['action'])
                return _show_cluster(senlin_client, parsed_args.cluster)
        else:
            print('Request error: %s' % resp)

        return '', ''


class ClusterNodeDel(command.ShowOne):
    """Delete specified nodes from cluster."""
    log = logging.getLogger(__name__ + ".ClusterNodeDel")

    def get_parser(self, prog_name):
        parser = super(ClusterNodeDel, self).get_parser(prog_name)
        parser.add_argument(
            '--nodes',
            metavar='<nodes>',
            required=True,
            help=_('Name or ID of nodes to be deleted; multiple nodes can be '
                   'separated with ","')
        )
        parser.add_argument(
            '-d',
            '--destroy-after-deletion',
            required=False,
            default=False,
            help=_('Whether nodes should be destroyed after deleted. '
                   'Default is False.')
        )
        parser.add_argument(
            'cluster',
            metavar='<cluster>',
            help=_('Name or ID of cluster to operate on')
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help=_('Wait for cluster members delete to complete')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering
        node_ids = parsed_args.nodes.split(',')
        destroy = parsed_args.destroy_after_deletion
        destroy = strutils.bool_from_string(destroy, strict=True)
        kwargs = {"destroy_after_deletion": destroy}
        resp = senlin_client.remove_nodes_from_cluster(
            parsed_args.cluster, node_ids, **kwargs)
        if 'action' in resp:
            print('Request accepted by action: %s' % resp['action'])
            if parsed_args.wait:
                senlin_utils.await_action(senlin_client, resp['action'])
                return _show_cluster(senlin_client, parsed_args.cluster)
        else:
            print('Request error: %s' % resp)

        return '', ''


class ClusterNodeReplace(command.ShowOne):
    """Replace the nodes in a cluster with specified nodes."""
    log = logging.getLogger(__name__ + ".ClusterNodeReplace")

    def get_parser(self, prog_name):
        parser = super(ClusterNodeReplace, self).get_parser(prog_name)
        parser.add_argument(
            '--nodes',
            metavar='<OLD_NODE1=NEW_NODE1>',
            required=True,
            help=_("OLD_NODE is the name or ID of a node to be replaced, "
                   "NEW_NODE is the name or ID of a node as replacement. "
                   "This can be specified multiple times, or once with "
                   "node-pairs separated by a comma ','."),
            action='append'
        )
        parser.add_argument(
            'cluster',
            metavar='<cluster>',
            help=_('Name or ID of cluster to operate on')
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help=_('Wait for cluster members replace to complete')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering
        nodepairs = {}
        for nodepair in parsed_args.nodes:
            key = nodepair.split('=')[0]
            value = nodepair.split('=')[1]
            nodepairs[key] = value
        resp = senlin_client.replace_nodes_in_cluster(parsed_args.cluster,
                                                      nodepairs)
        if 'action' in resp:
            print('Request accepted by action: %s' % resp['action'])
            if parsed_args.wait:
                senlin_utils.await_action(senlin_client, resp['action'])
                return _show_cluster(senlin_client, parsed_args.cluster)
        else:
            print('Request error: %s' % resp)

        return '', ''


class CheckCluster(command.Lister):
    """Check the cluster(s)."""
    log = logging.getLogger(__name__ + ".CheckCluster")

    def get_parser(self, prog_name):
        parser = super(CheckCluster, self).get_parser(prog_name)
        parser.add_argument(
            'cluster',
            metavar='<cluster>',
            nargs='+',
            help=_('ID or name of cluster(s) to operate on.')
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help=_('Wait for cluster check to complete')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering

        cluster_actions = {}

        for cid in parsed_args.cluster:
            try:
                resp = senlin_client.check_cluster(cid)
            except sdk_exc.ResourceNotFound:
                raise exc.CommandError(_('Cluster not found: %s') % cid)
            if 'action' in resp:
                print('Cluster check request on cluster %(cid)s is '
                      'accepted by action %(action)s.'
                      % {'cid': cid, 'action': resp['action']})
                cluster_actions[cid] = resp['action']
            else:
                print('Request error: %s' % resp)

        # generate the output after all actions have been accepted/rejected
        if parsed_args.wait and len(cluster_actions) > 0:
            for cid, action in cluster_actions.items():
                senlin_utils.await_action(senlin_client, action)
                senlin_utils.await_cluster_status(senlin_client, cid)
            return _list_cluster_summaries(senlin_client,
                                           cluster_actions.keys())

        return '', ''


def _list_cluster_summaries(senlin_client, cluster_ids):
    clusters = []
    for cluster_id in cluster_ids:
        try:
            cluster = senlin_client.get_cluster(cluster_id)
        except sdk_exc.ResourceNotFound:
            raise exc.CommandError(_('Cluster not found: %s') % cluster_id)

        clusters.append(cluster)

    columns = ['ID', 'Name', 'Status', 'Status Reason']
    formatters = {}
    props = (utils.get_item_properties(c, columns, formatters=formatters)
             for c in clusters)
    return columns, props


class RecoverCluster(command.Lister):
    """Recover the cluster(s)."""
    log = logging.getLogger(__name__ + ".RecoverCluster")

    def get_parser(self, prog_name):
        parser = super(RecoverCluster, self).get_parser(prog_name)
        parser.add_argument(
            'cluster',
            metavar='<cluster>',
            nargs='+',
            help=_('ID or name of cluster(s) to operate on.')
        )
        parser.add_argument(
            '--check',
            metavar='<boolean>',
            default=False,
            help=_("Whether the cluster should check it's nodes status before "
                   "doing cluster recover. Default is false")
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help=_('Wait for cluster recover to complete')
        )

        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering

        params = {
            'check': strutils.bool_from_string(parsed_args.check, strict=True)
        }

        cluster_actions = {}
        for cid in parsed_args.cluster:
            try:
                resp = senlin_client.recover_cluster(cid, **params)
            except sdk_exc.ResourceNotFound:
                raise exc.CommandError(_('Cluster not found: %s') % cid)
            if 'action' in resp:
                print('Cluster recover request on cluster %(cid)s is '
                      'accepted by action %(action)s.'
                      % {'cid': cid, 'action': resp['action']})
                cluster_actions[cid] = resp['action']
            else:
                print('Request error: %s' % resp)

        # generate the output after all actions have been accepted/rejected
        if parsed_args.wait and len(cluster_actions) > 0:
            for cid, action in cluster_actions.items():
                senlin_utils.await_action(senlin_client, action)
                senlin_utils.await_cluster_status(senlin_client, cid)
            return _list_cluster_summaries(senlin_client,
                                           cluster_actions.keys())

        return '', ''


class ClusterCollect(command.Lister):
    """Collect attributes across a cluster."""
    log = logging.getLogger(__name__ + ".ClusterCollect")

    def get_parser(self, prog_name):
        parser = super(ClusterCollect, self).get_parser(prog_name)
        parser.add_argument(
            '--full-id',
            default=False,
            action="store_true",
            help=_('Print full IDs in list')
        )
        parser.add_argument(
            '--path',
            metavar='<path>',
            required=True,
            help=_('JSON path expression for attribute to be collected')
        )
        parser.add_argument(
            'cluster',
            metavar='<cluster>',
            help=_('ID or name of cluster(s) to operate on.')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering
        attrs = senlin_client.collect_cluster_attrs(parsed_args.cluster,
                                                    parsed_args.path)
        columns = ['node_id', 'attr_value']
        formatters = {}
        if not parsed_args.full_id:
            formatters = {
                'node_id': lambda x: x[:8]
            }
        return (columns,
                (utils.get_item_properties(a, columns, formatters=formatters)
                 for a in attrs))


class ClusterOp(command.ShowOne):
    """Perform an operation on all nodes across a cluster."""
    log = logging.getLogger(__name__ + ".ClusterOp")

    def get_parser(self, prog_name):
        parser = super(ClusterOp, self).get_parser(prog_name)
        parser.add_argument(
            '--operation',
            metavar='<operation>',
            required=True,
            help=_('Operation to be performed on the cluster')
        )
        parser.add_argument(
            '--params',
            metavar='<key1=value1;key2=value2...>',
            help=_("Parameters to for the specified operation. "
                   "This can be specified multiple times, or once with "
                   "parameters separated by a semicolon."),
            action='append'
        )
        parser.add_argument(
            'cluster',
            metavar='<cluster>',
            help=_('ID or name of cluster to operate on.')
        )
        parser.add_argument(
            '--wait',
            action='store_true',
            help=_('Wait for cluster operation to complete')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering
        cid = parsed_args.cluster
        if parsed_args.params:
            params = senlin_utils.format_parameters(parsed_args.params)
        else:
            params = {}

        try:
            resp = senlin_client.perform_operation_on_cluster(
                cid, parsed_args.operation, **params)
        except sdk_exc.ResourceNotFound:
            raise exc.CommandError(_('Cluster not found: %s') % cid)
        if 'action' in resp:
            print('Request accepted by action: %s' % resp['action'])
            if parsed_args.wait:
                senlin_utils.await_action(senlin_client, resp['action'])
                senlin_utils.await_cluster_status(senlin_client, cid)
                return _show_cluster(senlin_client, cid)
        else:
            print('Request error: %s' % resp)

        return '', ''


class ClusterRun(command.Command):
    """Run scripts on cluster."""
    log = logging.getLogger(__name__ + ".ClusterRun")

    def get_parser(self, prog_name):
        parser = super(ClusterRun, self).get_parser(prog_name)
        parser.add_argument(
            '--port',
            metavar='<port>',
            type=int,
            default=22,
            help=_('The TCP port to use for SSH connection')
        )
        parser.add_argument(
            '--address-type',
            metavar='<address_type>',
            default='floating',
            help=_("The type of IP address to use. Possible values include "
                   "'fixed' and 'floating' (the default)")
        )
        parser.add_argument(
            '--network',
            metavar='<network>',
            default='',
            help=_("The network to use for SSH connection")
        )
        parser.add_argument(
            '--ipv6',
            action="store_true",
            default=False,
            help=_("Whether the IPv6 address should be used for SSH. Default "
                   "to use IPv4 address.")
        )
        parser.add_argument(
            '--user',
            metavar='<user>',
            default='root',
            help=_("The login name to use for SSH connection. Default to "
                   "'root'.")
        )
        parser.add_argument(
            '--identity-file',
            metavar='<identity_file>',
            help=_("The private key file to use, same as the '-i' SSH option")
        )
        parser.add_argument(
            '--ssh-options',
            metavar='<ssh_options>',
            default="",
            help=_("Extra options to pass to SSH. See: man ssh.")
        )
        parser.add_argument(
            '--script',
            metavar='<script>',
            required=True,
            help=_("Path name of the script file to run")
        )
        parser.add_argument(
            'cluster',
            metavar='<cluster>',
            help=_('ID or name of cluster(s) to operate on.')
        )
        return parser

    def take_action(self, args):
        self.log.debug("take_action(%s)", args)
        service = self.app.client_manager.clustering

        if '@' in args.cluster:
            user, cluster = args.cluster.split('@', 1)
            args.user = user
            args.cluster = cluster

        try:
            attributes = service.collect_cluster_attrs(args.cluster, 'details')
        except sdk_exc.ResourceNotFound:
            raise exc.CommandError(_("Cluster not found: %s") % args.cluster)

        try:
            f = open(args.script, 'r')
            script = f.read()
        except Exception:
            raise exc.CommandError(_("Could not open script file: %s") %
                                   args.script)

        tasks = dict()
        for attr in attributes:
            node_id = attr.node_id
            addr = attr.attr_value['addresses']

            output = dict()
            th = threading.Thread(
                target=self._run_script,
                args=(node_id, addr, args.network, args.address_type,
                      args.port, args.user, args.ipv6, args.identity_file,
                      script, args.ssh_options),
                kwargs={'output': output})
            th.start()
            tasks[th] = (node_id, output)

        for t in tasks:
            t.join()

        for t in tasks:
            node_id, result = tasks[t]
            print("node: %s" % node_id)
            print("status: %s" % result.get('status'))
            if "reason" in result:
                print("reason: %s" % result.get('reason'))
            if "output" in result:
                print("output:\n%s" % result.get('output'))
            if "error" in result:
                print("error:\n%s" % result.get('error'))

    def _run_script(self, node_id, addr, net, addr_type, port, user, ipv6,
                    identity_file, script, options, output=None):
        version = 6 if ipv6 else 4

        # Select the network to use.
        if net:
            addresses = addr.get(net)
            if not addresses:
                output['status'] = _('FAILED')
                output['error'] = _("Node '%(node)s' is not attached to "
                                    "network '%(net)s'.") % {'node': node_id,
                                                             'net': net}
                return
        else:
            # network not specified
            if len(addr) > 1:
                output['status'] = _('FAILED')
                output['error'] = _("Node '%(node)s' is attached to more "
                                    "than one network. Please pick the "
                                    "network to use.") % {'node': node_id}
                return
            elif not addr:
                output['status'] = _('FAILED')
                output['error'] = _("Node '%(node)s' is not attached to any "
                                    "network.") % {'node': node_id}
                return
            else:
                addresses = list(addr.values())[0]

        # Select the address in the selected network.
        # If the extension is not present, we assume the address to be
        # floating.
        matching_addresses = []
        for a in addresses:
            a_type = a.get('OS-EXT-IPS:type', 'floating')
            a_version = a.get('version')
            if (a_version == version and a_type == addr_type):
                matching_addresses.append(a.get('addr'))

        if not matching_addresses:
            output['status'] = _('FAILED')
            output['error'] = _("No address that matches network '%(net)s' "
                                "and type '%(type)s' of IPv%(ver)s has been "
                                "found for node '%(node)s'."
                                ) % {'net': net, 'type': addr_type,
                                     'ver': version, 'node': node_id}
            return

        if len(matching_addresses) > 1:
            output['status'] = _('FAILED')
            output['error'] = _("More than one IPv%(ver)s %(type)s address "
                                "found.") % {'ver': version,
                                             'type': addr_type}
            return

        ip_address = str(matching_addresses[0])
        identity = '-i %s' % identity_file if identity_file else ''

        cmd = [
            'ssh',
            '-%d' % version,
            '-p%d' % port,
            identity,
            options,
            '%s@%s' % (user, ip_address),
            '%s' % script
        ]

        self.log.debug("%s" % cmd)
        proc = subprocess.Popen(cmd, stdout=subprocess.PIPE)
        (stdout, stderr) = proc.communicate()

        while proc.returncode is None:
            time.sleep(1)

        if proc.returncode == 0:
            output['status'] = _('SUCCEEDED (0)')
            output['output'] = stdout
            if stderr:
                output['error'] = stderr
        else:
            output['status'] = _('FAILED (%d)') % proc.returncode
            output['output'] = stdout
            if stderr:
                output['error'] = stderr



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Clustering v1 event action implementations"""

import logging

from openstack import exceptions as sdk_exc
from osc_lib.command import command
from osc_lib import exceptions as exc
from osc_lib import utils

from senlinclient.common.i18n import _
from senlinclient.common import utils as senlin_utils


class ListEvent(command.Lister):
    """List events."""

    log = logging.getLogger(__name__ + ".ListEvent")

    def get_parser(self, prog_name):
        parser = super(ListEvent, self).get_parser(prog_name)
        parser.add_argument(
            '--filters',
            metavar='<"key1=value1;key2=value2...">',
            help=_("Filter parameters to apply on returned events. "
                   "This can be specified multiple times, or once with "
                   "parameters separated by a semicolon. The valid filter "
                   "keys are: ['level', 'otype', 'oid' ,'cluster_id', "
                   "'oname', 'action']. "
                   "NOTICE: The value of 'oid' or 'cluster_id', "
                   "if provided, must be a full ID."),
            action='append'
        )
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            help=_('Limit the number of events returned')
        )
        parser.add_argument(
            '--marker',
            metavar='<id>',
            help=_('Only return events that appear after the given event ID')
        )
        parser.add_argument(
            '--sort',
            metavar='<key>[:<direction>]',
            help=_("Sorting option which is a string containing a list of "
                   "keys separated by commas. Each key can be optionally "
                   "appended by a sort direction (:asc or :desc). The valid "
                   "sort keys are: ['timestamp', 'level', 'oid', 'otype', "
                   "'oname', 'action', 'status']")
        )
        parser.add_argument(
            '--global-project',
            default=False,
            action="store_true",
            help=_('Whether events from all projects should be listed. '
                   ' Default to False. Setting this to True may demand '
                   'for an admin privilege')
        )
        parser.add_argument(
            '--full-id',
            default=False,
            action="store_true",
            help=_('Print full IDs in list')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        senlin_client = self.app.client_manager.clustering
        columns = ['id', 'generated_at', 'obj_type', 'obj_id', 'obj_name',
                   'action', 'status', 'level', 'cluster_id', 'meta_data']
        queries = {
            'sort': parsed_args.sort,
            'limit': parsed_args.limit,
            'marker': parsed_args.marker,
            'global_project': parsed_args.global_project,
        }

        if parsed_args.filters:
            queries.update(senlin_utils.format_parameters(parsed_args.filters))

        formatters = {}
        if parsed_args.global_project:
            columns.append('project_id')
        if not parsed_args.full_id:
            formatters['id'] = lambda x: x[:8]
            formatters['obj_id'] = lambda x: x[:8] if x else ''
            if 'project_id' in columns:
                formatters['project_id'] = lambda x: x[:8]
            formatters['cluster_id'] = lambda x: x[:8] if x else ''

        events = senlin_client.events(**queries)
        return (columns,
                (utils.get_item_properties(e, columns,
                                           formatters=formatters)
                 for e in events))


class ShowEvent(command.ShowOne):
    """Show the event details."""

    log = logging.getLogger(__name__ + ".ShowEvent")

    def get_parser(self, prog_name):
        parser = super(ShowEvent, self).get_parser(prog_name)
        parser.add_argument(
            'event',
            metavar='<event>',
            help=_('ID of event to display details for')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        senlin_client = self.app.client_manager.clustering
        try:
            event = senlin_client.get_event(parsed_args.event)
        except sdk_exc.ResourceNotFound:
            raise exc.CommandError(_("Event not found: %s")
                                   % parsed_args.event)
        data = event.to_dict()
        columns = sorted(data.keys())
        return columns, utils.get_dict_properties(data, columns)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import io
from unittest import mock

from openstack import exceptions as sdk_exc
from osc_lib import exceptions as exc

from senlinclient.tests.unit.v1 import fakes
from senlinclient.v1 import node as osc_node


class TestNode(fakes.TestClusteringv1):
    def setUp(self):
        super(TestNode, self).setUp()
        self.mock_client = self.app.client_manager.clustering


class TestNodeList(TestNode):

    columns = ['id', 'name', 'index', 'status', 'cluster_id',
               'physical_id', 'profile_name', 'created_at', 'updated_at',
               'tainted']

    defaults = {
        'cluster_id': None,
        'global_project': False,
        'marker': None,
        'limit': None,
        'sort': None,
    }

    def setUp(self):
        super(TestNodeList, self).setUp()
        self.cmd = osc_node.ListNode(self.app, None)
        fake_node = mock.Mock(
            cluster_id=None,
            created_at="2015-02-27T04:39:21",
            data={},
            details={},
            domain=None,
            id="573aa1ba-bf45-49fd-907d-6b5d6e6adfd3",
            index=-1,
            init_at="2015-02-27T04:39:18",
            metadata={},
            physical_id="cc028275-d078-4729-bf3e-154b7359814b",
            profile_id="edc63d0a-2ca4-48fa-9854-27926da76a4a",
            profile_name="mystack",
            project_id="6e18cc2bdbeb48a5b3cad2dc499f6804",
            role=None,
            status="ACTIVE",
            status_reason="Creation succeeded",
            tainted=True,
            updated_at=None,
            user_id="5e5bf8027826429c96af157f68dc9072"
        )
        fake_node.name = "node00a"
        fake_node.to_dict = mock.Mock(return_value={})
        self.mock_client.nodes = mock.Mock(return_value=[fake_node])

    def test_node_list_defaults(self):
        arglist = []
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.nodes.assert_called_with(**self.defaults)
        self.assertEqual(self.columns, columns)

    def test_node_list_full_id(self):
        arglist = ['--full-id']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.nodes.assert_called_with(**self.defaults)
        self.assertEqual(self.columns, columns)

    def test_node_list_limit(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['limit'] = '3'
        arglist = ['--limit', '3']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.nodes.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)

    def test_node_list_sort(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['sort'] = 'name:asc'
        arglist = ['--sort', 'name:asc']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.nodes.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)

    def test_node_list_sort_invalid_key(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['sort'] = 'bad_key'
        arglist = ['--sort', 'bad_key']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.mock_client.nodes.side_effect = sdk_exc.HttpException()
        self.assertRaises(sdk_exc.HttpException,
                          self.cmd.take_action, parsed_args)

    def test_node_list_sort_invalid_direction(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['sort'] = 'name:bad_direction'
        arglist = ['--sort', 'name:bad_direction']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.mock_client.nodes.side_effect = sdk_exc.HttpException()
        self.assertRaises(sdk_exc.HttpException,
                          self.cmd.take_action, parsed_args)

    def test_node_list_filter(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['name'] = 'my_node'
        arglist = ['--filter', 'name=my_node']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.nodes.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)

    def test_node_list_marker(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['marker'] = 'a9448bf6'
        arglist = ['--marker', 'a9448bf6']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.nodes.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)


class TestNodeShow(TestNode):
    response = {"node": {
    }}

    def setUp(self):
        super(TestNodeShow, self).setUp()
        self.cmd = osc_node.ShowNode(self.app, None)
        fake_node = mock.Mock(
            cluster_id=None,
            created_at="2015-02-10T12:03:16",
            data={},
            details={"OS-DCF:diskConfig": "MANUAL"},
            domain_id=None,
            id="d5779bb0-f0a0-49c9-88cc-6f078adb5a0b",
            index=-1,
            init_at="2015-02-10T12:03:13",
            metadata={},
            physical_id="f41537fa-22ab-4bea-94c0-c874e19d0c80",
            profile_id="edc63d0a-2ca4-48fa-9854-27926da76a4a",
            profile_name="mystack",
            project_id="6e18cc2bdbeb48a5b3cad2dc499f6804",
            role=None,
            status="ACTIVE",
            status_reason="Creation succeeded",
            tainted=True,
            updated_at="2015-03-04T04:58:27",
            user_id="5e5bf8027826429c96af157f68dc9072"
        )
        fake_node.name = "my_node"
        fake_node.to_dict = mock.Mock(
            return_value={'details': {'key': 'value'}}
        )
        self.mock_client.get_node = mock.Mock(return_value=fake_node)

    def test_node_show(self):
        arglist = ['my_node']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.get_node.assert_called_with('my_node', details=False)

    def test_node_show_with_details(self):
        arglist = ['my_node', '--details']
        parsed_args = self.check_parser(self.cmd, arglist, [])

        self.cmd.take_action(parsed_args)

        self.mock_client.get_node.assert_called_with(
            'my_node', details=True)

    def test_node_show_not_found(self):
        arglist = ['my_node']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.mock_client.get_node.side_effect = sdk_exc.ResourceNotFound()
        error = self.assertRaises(exc.CommandError, self.cmd.take_action,
                                  parsed_args)
        self.assertEqual('Node not found: my_node', str(error))


class TestNodeCreate(TestNode):

    defaults = {
        "cluster_id": None,
        "metadata": {},
        "name": "my_node",
        "profile_id": "mystack",
        "role": None
    }

    def setUp(self):
        super(TestNodeCreate, self).setUp()
        self.cmd = osc_node.CreateNode(self.app, None)
        fake_node = mock.Mock(
            action="2366d440-c73e-4961-9254-6d1c3af7c167",
            cluster_id="",
            created_at=None,
            data={},
            domain=None,
            id="0df0931b-e251-4f2e-8719-4ebfda3627ba",
            index=-1,
            init_time="2015-03-05T08:53:15",
            metadata={},
            physical_id=None,
            profile_id="edc63d0a-2ca4-48fa-9854-27926da76a4a",
            profile_name="mystack",
            project_id="6e18cc2bdbeb48a5b3cad2dc499f6804",
            role="master",
            status="INIT",
            status_reason="Initializing",
            updated_at=None,
            user_id="5e5bf8027826429c96af157f68dc9072"
        )
        fake_node.name = "my_node"
        fake_node.to_dict = mock.Mock(return_value={})

        self.mock_client.create_node = mock.Mock(return_value=fake_node)
        self.mock_client.get_node = mock.Mock(return_value=fake_node)

    def test_node_create_defaults(self):
        arglist = ['my_node', '--profile', 'mystack']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.create_node.assert_called_with(**self.defaults)

    def test_node_create_with_metadata(self):
        arglist = ['my_node', '--profile', 'mystack',
                   '--metadata', 'key1=value1;key2=value2']
        kwargs = copy.deepcopy(self.defaults)
        kwargs['metadata'] = {'key1': 'value1', 'key2': 'value2'}
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.create_node.assert_called_with(**kwargs)

    def test_node_create_with_cluster(self):
        arglist = ['my_node', '--profile', 'mystack',
                   '--cluster', 'mycluster']
        kwargs = copy.deepcopy(self.defaults)
        kwargs['cluster_id'] = 'mycluster'
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.create_node.assert_called_with(**kwargs)

    def test_node_create_with_role(self):
        arglist = ['my_node', '--profile', 'mystack',
                   '--role', 'master']
        kwargs = copy.deepcopy(self.defaults)
        kwargs['role'] = 'master'
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.create_node.assert_called_with(**kwargs)


class TestNodeUpdate(TestNode):

    defaults = {
        "name": "new_node",
        "metadata": {
            "nk1": "nv1",
            "nk2": "nv2",
        },
        "profile_id": "new_profile",
        "role": "new_role",
        "tainted": True
    }

    def setUp(self):
        super(TestNodeUpdate, self).setUp()
        self.cmd = osc_node.UpdateNode(self.app, None)
        fake_node = mock.Mock(
            action="2366d440-c73e-4961-9254-6d1c3af7c167",
            cluster_id="",
            created_at=None,
            data={},
            domain=None,
            id="0df0931b-e251-4f2e-8719-4ebfda3627ba",
            index=-1,
            init_time="2015-03-05T08:53:15",
            metadata={},
            physical_id=None,
            profile_id="edc63d0a-2ca4-48fa-9854-27926da76a4a",
            profile_name="mystack",
            project_id="6e18cc2bdbeb48a5b3cad2dc499f6804",
            role="master",
            status="INIT",
            status_reason="Initializing",
            tainted=False,
            updated_at=None,
            user_id="5e5bf8027826429c96af157f68dc9072"
        )
        fake_node.name = "my_node"
        fake_node.to_dict = mock.Mock(return_value={})
        self.mock_client.update_node = mock.Mock(return_value=fake_node)
        self.mock_client.get_node = mock.Mock(return_value=fake_node)
        self.mock_client.find_node = mock.Mock(return_value=fake_node)

    def test_node_update_defaults(self):
        arglist = ['--name', 'new_node', '--metadata', 'nk1=nv1;nk2=nv2',
                   '--profile', 'new_profile', '--role', 'new_role',
                   '--tainted', 'True', '0df0931b']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.update_node.assert_called_with(
            '0df0931b-e251-4f2e-8719-4ebfda3627ba', **self.defaults)

    def test_node_update_not_found(self):
        arglist = ['--name', 'new_node', '--metadata', 'nk1=nv1;nk2=nv2',
                   '--profile', 'new_profile', '--role', 'new_role',
                   'c6b8b252']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.mock_client.find_node.return_value = None
        error = self.assertRaises(exc.CommandError, self.cmd.take_action,
                                  parsed_args)
        self.assertIn('Node not found: c6b8b252', str(error))


class TestNodeDelete(TestNode):
    def setUp(self):
        super(TestNodeDelete, self).setUp()
        self.cmd = osc_node.DeleteNode(self.app, None)
        mock_node = mock.Mock(location='loc/fake_action_id')
        self.mock_client.delete_node = mock.Mock(return_value=mock_node)

    def test_node_delete(self):
        arglist = ['node1', 'node2', 'node3']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.delete_node.assert_has_calls(
            [mock.call('node1', False, False),
             mock.call('node2', False, False),
             mock.call('node3', False, False)]
        )

    def test_node_delete_force(self):
        arglist = ['node1', 'node2', 'node3', '--force']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.delete_node.assert_has_calls(
            [mock.call('node1', False, False),
             mock.call('node2', False, False),
             mock.call('node3', False, False)]
        )

    def test_node_delete_force_delete(self):
        arglist = ['node1', 'node2', 'node3', '--force-delete']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.delete_node.assert_has_calls(
            [mock.call('node1', False, True),
             mock.call('node2', False, True),
             mock.call('node3', False, True)]
        )

    def test_node_delete_not_found(self):
        arglist = ['my_node']
        self.mock_client.delete_node.side_effect = sdk_exc.ResourceNotFound
        parsed_args = self.check_parser(self.cmd, arglist, [])

        self.cmd.take_action(parsed_args)

        self.mock_client.delete_node.assert_has_calls(
            [mock.call('my_node', False, False)]
        )

    def test_node_delete_one_found_one_not_found(self):
        arglist = ['node1', 'node2']
        self.mock_client.delete_node.side_effect = (
            [None, sdk_exc.ResourceNotFound]
        )
        parsed_args = self.check_parser(self.cmd, arglist, [])

        self.cmd.take_action(parsed_args)

        self.mock_client.delete_node.assert_has_calls(
            [mock.call('node1', False, False),
             mock.call('node2', False, False)]
        )

    @mock.patch('sys.stdin', spec=io.StringIO)
    def test_node_delete_prompt_yes(self, mock_stdin):
        arglist = ['my_node']
        mock_stdin.isatty.return_value = True
        mock_stdin.readline.return_value = 'y'
        parsed_args = self.check_parser(self.cmd, arglist, [])

        self.cmd.take_action(parsed_args)

        mock_stdin.readline.assert_called_with()
        self.mock_client.delete_node.assert_called_with(
            'my_node', False, False)

    @mock.patch('sys.stdin', spec=io.StringIO)
    def test_node_delete_prompt_no(self, mock_stdin):
        arglist = ['my_node']
        mock_stdin.isatty.return_value = True
        mock_stdin.readline.return_value = 'n'
        parsed_args = self.check_parser(self.cmd, arglist, [])

        self.cmd.take_action(parsed_args)

        mock_stdin.readline.assert_called_with()
        self.mock_client.delete_node.assert_not_called()


class TestNodeCheck(TestNode):
    response = {"action": "8bb476c3-0f4c-44ee-9f64-c7b0260814de"}

    def setUp(self):
        super(TestNodeCheck, self).setUp()
        self.cmd = osc_node.CheckNode(self.app, None)
        self.mock_client.check_node = mock.Mock(
            return_value=self.response)

    def test_node_check(self):
        arglist = ['node1', 'node2', 'node3']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.check_node.assert_has_calls(
            [mock.call('node1'), mock.call('node2'),
             mock.call('node3')]
        )

    def test_node_check_not_found(self):
        arglist = ['node1']
        self.mock_client.check_node.side_effect = sdk_exc.ResourceNotFound
        parsed_args = self.check_parser(self.cmd, arglist, [])
        error = self.assertRaises(exc.CommandError, self.cmd.take_action,
                                  parsed_args)
        self.assertIn('Node not found: node1', str(error))


class TestNodeRecover(TestNode):

    action_id = "8bb476c3-0f4c-44ee-9f64-c7b0260814de"

    def setUp(self):
        super(TestNodeRecover, self).setUp()
        self.cmd = osc_node.RecoverNode(self.app, None)
        fake_res = {'action': self.action_id}
        self.mock_client.recover_node = mock.Mock(return_value=fake_res)

    def test_node_recover(self):
        arglist = ['node1', 'node2', 'node3', '--check', 'false']
        kwargs = {'check': False}
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.recover_node.assert_has_calls(
            [mock.call('node1', **kwargs), mock.call('node2', **kwargs),
             mock.call('node3', **kwargs)]
        )

    def test_node_recover_not_found(self):
        arglist = ['node1']
        self.mock_client.recover_node.side_effect = sdk_exc.ResourceNotFound
        parsed_args = self.check_parser(self.cmd, arglist, [])
        error = self.assertRaises(exc.CommandError, self.cmd.take_action,
                                  parsed_args)
        self.assertIn('Node not found: node1', str(error))


class TestNodeAdopt(TestNode):
    defaults = {
        "identity": "fake-resource-id",
        "metadata": {},
        "name": "my_node",
        "overrides": {},
        "role": None,
        "snapshot": False,
        "type": "os.nova.server-1.0"
    }

    def setUp(self):
        super(TestNodeAdopt, self).setUp()
        self.cmd = osc_node.AdoptNode(self.app, None)
        fake_node = mock.Mock(
            action="2366d440-c73e-4961-9254-6d1c3af7c167",
            cluster_id="",
            created_at=None,
            data={},
            domain=None,
            id="0df0931b-e251-4f2e-8719-4ebfda3627ba",
            index=-1,
            init_time="2015-03-05T08:53:15",
            metadata={},
            physical_id=None,
            profile_id="edc63d0a-2ca4-48fa-9854-27926da76a4a",
            profile_name="mystack",
            project_id="6e18cc2bdbeb48a5b3cad2dc499f6804",
            role="master",
            status="INIT",
            status_reason="Initializing",
            updated_at=None,
            user_id="5e5bf8027826429c96af157f68dc9072"
        )
        fake_node.name = "my_node"
        fake_node.to_dict = mock.Mock(return_value={})

        self.mock_client.adopt_node = mock.Mock(return_value=fake_node)
        self.mock_client.get_node = mock.Mock(return_value=fake_node)

    def test_node_adopt_defaults(self):
        arglist = ['--identity', 'fake-resource-id',
                   '--type', 'os.nova.server-1.0',
                   '--name', 'my_node']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.adopt_node.assert_called_with(False, **self.defaults)

    def test_node_adopt_with_metadata(self):
        arglist = ['--identity', 'fake-resource-id',
                   '--type', 'os.nova.server-1.0',
                   '--metadata', 'key1=value1;key2=value2',
                   '--name', 'my_node']
        kwargs = copy.deepcopy(self.defaults)
        kwargs['metadata'] = {'key1': 'value1', 'key2': 'value2'}
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.adopt_node.assert_called_with(False, **kwargs)

    def test_node_adopt_with_override(self):
        arglist = ['--identity', 'fake-resource-id',
                   '--type', 'os.nova.server-1.0',
                   '--overrides',
                   '{"networks": [{"network": "fake-net-name"}]}',
                   '--name', 'my_node']
        kwargs = copy.deepcopy(self.defaults)
        kwargs['overrides'] = {'networks': [{'network': 'fake-net-name'}]}
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.adopt_node.assert_called_with(False, **kwargs)

    def test_node_adopt_with_role(self):
        arglist = ['--identity', 'fake-resource-id',
                   '--type', 'os.nova.server-1.0',
                   '--role', 'master',
                   '--name', 'my_node']
        kwargs = copy.deepcopy(self.defaults)
        kwargs['role'] = 'master'
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.adopt_node.assert_called_with(False, **kwargs)

    def test_node_adopt_with_snapshot(self):
        arglist = ['--identity', 'fake-resource-id',
                   '--type', 'os.nova.server-1.0',
                   '--snapshot',
                   '--name', 'my_node']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        kwargs = copy.deepcopy(self.defaults)
        kwargs['snapshot'] = True
        self.cmd.take_action(parsed_args)
        self.mock_client.adopt_node.assert_called_with(False, **kwargs)


class TestNodeAdoptPreview(TestNode):
    defaults = {
        "identity": "fake-resource-id",
        "overrides": {},
        "snapshot": False,
        "type": "os.nova.server-1.0"
    }

    def setUp(self):
        super(TestNodeAdoptPreview, self).setUp()
        self.cmd = osc_node.AdoptNode(self.app, None)
        self.fake_node_preview = {
            "node_profile": {
                "node_preview": {
                    "properties": {

                    },
                    "type": "os.nova.server",
                    "version": "1.0"}
            }
        }

        self.mock_client.adopt_node = mock.Mock(
            return_value=self.fake_node_preview)
        self.mock_client.get_node = mock.Mock(
            return_value=self.fake_node_preview)

    def test_node_adopt_preview_default(self):
        arglist = ['--identity', 'fake-resource-id',
                   '--type', 'os.nova.server-1.0',
                   '--preview']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.adopt_node.assert_called_with(True, **self.defaults)

    def test_node_adopt_preview_with_overrides(self):
        arglist = ['--identity', 'fake-resource-id',
                   '--type', 'os.nova.server-1.0',
                   '--overrides',
                   '{"networks": [{"network": "fake-net-name"}]}',
                   '--preview']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        kwargs = copy.deepcopy(self.defaults)
        kwargs['overrides'] = {'networks': [{'network': 'fake-net-name'}]}
        self.cmd.take_action(parsed_args)
        self.mock_client.adopt_node.assert_called_with(True, **kwargs)

    def test_node_adopt_preview_with_snapshot(self):
        arglist = ['--identity', 'fake-resource-id',
                   '--type', 'os.nova.server-1.0',
                   '--snapshot', '--preview']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        kwargs = copy.deepcopy(self.defaults)
        kwargs['snapshot'] = True
        self.cmd.take_action(parsed_args)
        self.mock_client.adopt_node.assert_called_with(True, **kwargs)


class TestNodeOp(TestNode):

    response = {"action": "1db0f5c5-9183-4c47-9ef1-a5a97402a2c1"}

    def setUp(self):
        super(TestNodeOp, self).setUp()
        self.cmd = osc_node.NodeOp(self.app, None)
        self.mock_client.perform_operation_on_node = mock.Mock(
            return_value=self.response)

    def test_node_op(self):
        arglist = ['--operation', 'dance', '--params', 'style=tango',
                   'my_node']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.perform_operation_on_node.assert_called_once_with(
            'my_node',
            'dance',
            style='tango')

    def test_node_op_not_found(self):
        arglist = ['--operation', 'dance', 'node1']
        ex = sdk_exc.ResourceNotFound
        self.mock_client.perform_operation_on_node.side_effect = ex
        parsed_args = self.check_parser(self.cmd, arglist, [])
        error = self.assertRaises(exc.CommandError, self.cmd.take_action,
                                  parsed_args)
        self.assertIn('Node not found: node1', str(error))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from senlinclient.common import utils


def do_command_foo(sc, args):
    """Pydoc for command foo."""
    return


@utils.arg('-F', '--flag', metavar='<FLAG>', help='Flag desc.')
@utils.arg('arg1', metavar='<ARG1>', help='Arg1 desc')
def do_command_bar(sc, args):
    """This is the command doc."""
    return



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from senlinclient.tests.functional import base


class ProfileTest(base.OpenStackClientTestBase):
    """Test for profiles"""

    def test_profile_list(self):
        result = self.openstack('cluster profile list')
        profile_list = self.parser.listing(result)
        self.assertTableStruct(profile_list, ['id', 'name', 'type',
                                              'created_at'])

    def test_pofile_create(self):
        name = self.name_generate()
        result = self.profile_create(name, 'cirros_basic.yaml')
        self.assertEqual(result['name'], name)
        self.addCleanup(self.profile_delete, result['id'])

    def test_profile_update(self):
        old_name = self.name_generate()
        pf1 = self.profile_create(old_name, 'cirros_basic.yaml')
        new_name = self.name_generate()
        cmd = ('cluster profile update --name %s %s' % (new_name, pf1['id']))
        result = self.openstack(cmd)
        pf2 = self.show_to_dict(result)
        self.assertEqual(pf2['name'], new_name)
        self.assertNotEqual(pf1['name'], pf2['name'])
        self.addCleanup(self.profile_delete, pf2['id'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from senlinclient.tests.functional import base


class ClusterTest(base.OpenStackClientTestBase):
    """Test for clusters"""

    def test_cluster_list(self):
        result = self.openstack('cluster list')
        cluster_list = self.parser.listing(result)
        self.assertTableStruct(cluster_list, ['id', 'name', 'status',
                                              'created_at', 'updated_at'])

    def test_cluster_list_filters(self):
        params = {'name': 'test1', 'status': 'ACTIVE'}
        for k, v in params.items():
            cmd = ('cluster list --filters %s=%s' % (k, v))
            self.openstack(cmd)

    def test_cluster_list_sort(self):
        params = ['name', 'status', 'init_at', 'created_at', 'updated_at']
        for param in params:
            cmd = 'cluster list --sort %s' % param
            self.openstack(cmd)

    def test_cluster_full_id(self):
        self.openstack('cluster list --full-id')

    def test_cluster_limit(self):
        self.openstack('cluster list --limit 1')

    def test_cluster_create(self):
        name = self.name_generate()
        pf = self.profile_create(name)
        self.addCleanup(self.profile_delete, pf['id'])
        cluster = self.cluster_create(pf['id'], name, 1)
        self.addCleanup(self.cluster_delete, cluster['id'])
        cluster_raw = self.openstack('cluster show %s' % name)
        cluster_data = self.show_to_dict(cluster_raw)
        self.assertEqual(cluster_data['name'], name)
        self.assertEqual(cluster_data['status'], 'ACTIVE')
        self.assertEqual(cluster_data['desired_capacity'], '1')

    def test_cluster_update(self):
        old_name = self.name_generate()
        old_pf = self.profile_create(old_name)
        self.addCleanup(self.profile_delete, old_pf['id'])
        new_name = self.name_generate()
        new_pf = self.profile_create(new_name)
        self.addCleanup(self.profile_delete, new_pf['id'])
        cluster = self.cluster_create(old_pf['id'], old_name, 1)
        self.addCleanup(self.cluster_delete, cluster['id'])
        self.assertEqual(cluster['name'], old_name)

        # cluster update
        cmd = ('cluster update --name %s --profile %s --timeout 300 %s'
               % (old_name, new_pf['id'], cluster['id']))
        self.openstack(cmd)
        self.wait_for_status(cluster['id'], 'ACTIVE', 'cluster', 120)

        cluster_raw = self.openstack('cluster show %s' % cluster['id'])
        cluster_data = self.show_to_dict(cluster_raw)
        node_raw = self.openstack('cluster node show %s' %
                                  cluster_data['node_ids'])
        node_data = self.show_to_dict(node_raw)

        # if not profile-only, change all profile
        self.assertEqual(cluster['name'], cluster_data['name'])
        self.assertEqual(cluster_data['profile_id'], new_pf['id'])
        self.assertEqual(cluster_data['timeout'], '300')
        self.assertEqual(new_pf['name'], node_data['profile_name'])

    def test_cluster_update_profile_only(self):
        old_name = self.name_generate()
        old_pf = self.profile_create(old_name)
        self.addCleanup(self.profile_delete, old_pf['id'])
        new_name = self.name_generate()
        new_pf = self.profile_create(new_name)
        self.addCleanup(self.profile_delete, new_pf['id'])
        cluster = self.cluster_create(old_pf['id'], old_name, 1)
        self.addCleanup(self.cluster_delete, cluster['id'])
        self.assertEqual(cluster['name'], old_name)

        cmd = ('cluster update --name %s --profile %s --profile-only true'
               ' --timeout 300 %s' % (new_name, new_pf['id'], cluster['id']))
        self.openstack(cmd)
        self.wait_for_status(cluster['id'], 'ACTIVE', 'cluster', 120)

        cluster_raw = self.openstack('cluster show %s' % cluster['id'])
        cluster_data = self.show_to_dict(cluster_raw)
        node_raw = self.openstack('cluster node show %s' %
                                  cluster_data['node_ids'])
        node_data = self.show_to_dict(node_raw)

        # if profile-only true, not change exist node profile
        self.assertNotEqual(cluster['name'], cluster_data['name'])
        self.assertNotEqual(cluster_data['profile_id'], cluster['profile_id'])
        self.assertEqual(cluster_data['profile_id'], new_pf['id'])
        self.assertEqual(cluster_data['timeout'], '300')
        self.assertNotEqual(new_name, node_data['profile_name'])

    def test_cluster_show(self):
        name = self.name_generate()
        pf = self.profile_create(name)
        self.addCleanup(self.profile_delete, pf['id'])
        cluster = self.cluster_create(pf['id'], name)
        self.addCleanup(self.cluster_delete, cluster['id'])
        cluster_raw = self.openstack('cluster show %s' % name)
        cluster_data = self.show_to_dict(cluster_raw)
        self.assertIn('node_ids', cluster_data)
        self.assertIn('timeout', cluster_data)

    def test_cluster_expand_and_shrink(self):
        name = self.name_generate()
        pf = self.profile_create(name)
        self.addCleanup(self.profile_delete, pf['id'])
        cluster = self.cluster_create(pf['id'], name)
        self.addCleanup(self.cluster_delete, cluster['id'])
        cluster_raw = self.openstack('cluster show %s' % name)
        cluster_data = self.show_to_dict(cluster_raw)

        # cluster expand
        self.openstack('cluster expand --count 1 %s' % name)
        self.wait_for_status(cluster['id'], 'ACTIVE', 'cluster', 120)
        expand_raw = self.openstack('cluster show %s' % name)
        expand_data = self.show_to_dict(expand_raw)
        self.assertNotEqual(cluster_data['desired_capacity'],
                            expand_data['desired_capacity'])
        self.assertEqual(expand_data['desired_capacity'], '1')

        # cluster shrink
        self.openstack('cluster shrink --count 1 %s' % name)
        self.wait_for_status(cluster['id'], 'ACTIVE', 'cluster', 120)
        shrink_raw = self.openstack('cluster show %s' % name)
        shrink_data = self.show_to_dict(shrink_raw)
        self.assertNotEqual(shrink_data['desired_capacity'],
                            expand_data['desired_capacity'])
        self.assertEqual(cluster_data['desired_capacity'],
                         shrink_data['desired_capacity'])

    # NOTE(chenyb4): Since functional tests only focus on the client/server
    # interaction without invovling other OpenStack services, it is not
    # possible to mock a cluster failure and then test if the check logic
    # works. Such tests would be left to integration tests instead.
    def test_cluster_check(self):
        name = self.name_generate()
        pf = self.profile_create(name)
        self.addCleanup(self.profile_delete, pf['id'])
        cluster = self.cluster_create(pf['id'], name, 1)
        self.addCleanup(self.cluster_delete, cluster['id'])
        self.openstack('cluster check %s' % cluster['id'])
        self.wait_for_status(cluster['id'], 'ACTIVE', 'cluster', 120)
        check_raw = self.openstack('cluster show %s' % name)
        check_data = self.show_to_dict(check_raw)
        self.assertIn('CLUSTER_CHECK', check_data['status_reason'])
        cluster_status = ['ACTIVE', 'WARNING']
        self.assertIn(check_data['status'], cluster_status)

    # NOTE(chenyb4): A end-to-end test of the cluster recover operation needs
    # to be done with other OpenStack services involved, thus out of scope
    # for functional tests. Such tests would be left to integration tests
    # instead.
    def test_cluster_recover(self):
        name = self.name_generate()
        pf = self.profile_create(name)
        self.addCleanup(self.profile_delete, pf['id'])
        cluster = self.cluster_create(pf['id'], name, 1)
        self.addCleanup(self.cluster_delete, cluster['id'])
        cmd = ('cluster recover --check true %s' % cluster['id'])
        self.openstack(cmd)
        self.wait_for_status(cluster['id'], 'ACTIVE', 'cluster', 120)
        recover_raw = self.openstack('cluster show %s' % name)
        recover_data = self.show_to_dict(recover_raw)
        self.assertIn('CLUSTER_RECOVER', recover_data['status_reason'])
        self.assertEqual('ACTIVE', recover_data['status'])

    def test_cluster_resize(self):
        name = self.name_generate()
        pf = self.profile_create(name)
        self.addCleanup(self.profile_delete, pf['id'])
        cluster = self.cluster_create(pf['id'], name)
        self.addCleanup(self.cluster_delete, cluster['id'])
        cluster_raw = self.openstack('cluster show %s' % name)
        cluster_data = self.show_to_dict(cluster_raw)
        self.assertEqual(cluster_data['desired_capacity'], '0')
        self.assertEqual(cluster_data['max_size'], '-1')
        self.assertEqual(cluster_data['min_size'], '0')
        cmd = ('cluster resize --max-size 5 --min-size 1 --adjustment 2 %s'
               % cluster['id'])
        self.openstack(cmd)
        self.wait_for_status(cluster['id'], 'ACTIVE', 'cluster', 120)
        resize_raw = self.openstack('cluster show %s' % name)
        resize_data = self.show_to_dict(resize_raw)
        self.assertEqual(resize_data['desired_capacity'], '2')
        self.assertEqual(resize_data['max_size'], '5')
        self.assertEqual(resize_data['min_size'], '1')

    def test_cluster_members_list(self):
        name = self.name_generate()
        pf = self.profile_create(name)
        self.addCleanup(self.profile_delete, pf['id'])
        cluster = self.cluster_create(pf['id'], name)
        self.addCleanup(self.cluster_delete, cluster['id'])
        result = self.openstack('cluster members list --full-id %s'
                                % cluster['id'])
        members_list = self.parser.listing(result)
        self.assertTableStruct(members_list, ['id', 'name', 'index',
                                              'status', 'physical_id',
                                              'created_at'])

    def test_cluster_members_add_and_del(self):
        name = self.name_generate()
        pf = self.profile_create(name)
        self.addCleanup(self.profile_delete, pf['id'])
        cluster = self.cluster_create(pf['id'], name)
        self.addCleanup(self.cluster_delete, cluster['name'])
        node = self.node_create(pf['id'], name)
        self.addCleanup(self.node_delete, node['id'])
        cluster_raw = self.openstack('cluster show %s' % name)
        cluster_data = self.show_to_dict(cluster_raw)
        self.assertEqual('', cluster_data['node_ids'])

        # Add exist node to cluster
        cmd = ('cluster members add --nodes %s %s' % (node['name'],
                                                      cluster['id']))
        self.openstack(cmd)
        self.wait_for_status(cluster['id'], 'ACTIVE', 'cluster', 120)

        mem_ad_raw = self.openstack('cluster show %s' % name)
        mem_ad_data = self.show_to_dict(mem_ad_raw)
        self.assertNotEqual('', mem_ad_data['node_ids'])
        self.assertIn(node['id'], mem_ad_data['node_ids'])

        # Delete node from cluster
        cmd = ('cluster members del --nodes %s %s' % (node['name'],
                                                      cluster['id']))
        self.openstack(cmd)
        self.wait_for_status(cluster['id'], 'ACTIVE', 'cluster', 120)
        mem_del_raw = self.openstack('cluster show %s' % name)
        mem_del_data = self.show_to_dict(mem_del_raw)
        self.assertEqual('', mem_del_data['node_ids'])
        self.assertNotIn(node['id'], mem_del_data['node_ids'])

    def test_cluster_members_replace(self):
        name = self.name_generate()
        pf = self.profile_create(name)
        self.addCleanup(self.profile_delete, pf['id'])
        cluster = self.cluster_create(pf['id'], name, 1)
        self.addCleanup(self.cluster_delete, cluster['id'])
        cluster_raw = self.openstack('cluster show %s' % name)
        cluster_data = self.show_to_dict(cluster_raw)

        # Create replace node
        new_node = self.node_create(pf['id'], name)
        self.addCleanup(self.node_delete, new_node['id'])
        self.assertNotIn(new_node['id'], cluster_data['node_ids'])

        # Cluster node replace
        old_node = cluster_data['node_ids']
        self.addCleanup(self.node_delete, old_node)
        cmd = ('cluster members replace --nodes %s=%s %s'
               % (old_node, new_node['id'], cluster['id']))
        self.openstack(cmd, flags='--debug')
        self.wait_for_status(cluster['id'], 'ACTIVE', 'cluster', 120)
        replace_raw = self.openstack('cluster show %s' % name)
        replace_data = self.show_to_dict(replace_raw)
        self.assertIn(new_node['id'], replace_data['node_ids'])
        self.assertNotIn(old_node, replace_data['node_ids'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from senlinclient.tests.functional import base


class EventTest(base.OpenStackClientTestBase):
    """Test for events"""

    def test_event_list(self):
        result = self.openstack('cluster event list')
        event_list = self.parser.listing(result)
        self.assertTableStruct(event_list, ['id', 'generated_at', 'obj_type',
                                            'obj_id', 'obj_name', 'action',
                                            'status', 'level', 'cluster_id',
                                            'meta_data'])



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Clustering v1 policy action implementations"""

import logging
import sys

from openstack import exceptions as sdk_exc
from osc_lib.command import command
from osc_lib import exceptions as exc
from osc_lib import utils

from senlinclient.common.i18n import _
from senlinclient.common import utils as senlin_utils


class ListPolicy(command.Lister):
    """List policies that meet the criteria."""

    log = logging.getLogger(__name__ + ".ListPolicy")

    def get_parser(self, prog_name):
        parser = super(ListPolicy, self).get_parser(prog_name)
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            help=_('Limit the number of policies returned')
        )
        parser.add_argument(
            '--marker',
            metavar='<id>',
            help=_('Only return policies that appear after the given policy '
                   'ID')
        )
        parser.add_argument(
            '--sort',
            metavar='<key>[:<direction>]',
            help=_("Sorting option which is a string containing a list of "
                   "keys separated by commas. Each key can be optionally "
                   "appended by a sort direction (:asc or :desc). The valid "
                   "sort keys are: ['type', 'name', 'created_at', "
                   "'updated_at']")
        )
        parser.add_argument(
            '--filters',
            metavar='<"key1=value1;key2=value2...">',
            help=_("Filter parameters to apply on returned policies. "
                   "This can be specified multiple times, or once with "
                   "parameters separated by a semicolon. The valid filter "
                   "keys are: ['type', 'name']"),
            action='append'
        )
        parser.add_argument(
            '--global-project',
            default=False,
            action="store_true",
            help=_('Indicate that the list should include policies from'
                   ' all projects. This option is subject to access policy '
                   'checking. Default is False')
        )
        parser.add_argument(
            '--full-id',
            default=False,
            action="store_true",
            help=_('Print full IDs in list')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        senlin_client = self.app.client_manager.clustering
        columns = ['id', 'name', 'type', 'created_at']
        queries = {
            'limit': parsed_args.limit,
            'marker': parsed_args.marker,
            'sort': parsed_args.sort,
            'global_project': parsed_args.global_project,
        }
        if parsed_args.filters:
            queries.update(senlin_utils.format_parameters(parsed_args.filters))

        policies = senlin_client.policies(**queries)
        formatters = {}
        if parsed_args.global_project:
            columns.append('project_id')
        if not parsed_args.full_id:
            formatters = {
                'id': lambda x: x[:8]
            }
            if 'project_id' in columns:
                formatters['project_id'] = lambda x: x[:8]

        return (
            columns,
            (utils.get_item_properties(p, columns, formatters=formatters)
             for p in policies)
        )


class ShowPolicy(command.ShowOne):
    """Show the policy details."""

    log = logging.getLogger(__name__ + ".ShowPolicy")

    def get_parser(self, prog_name):
        parser = super(ShowPolicy, self).get_parser(prog_name)
        parser.add_argument(
            'policy',
            metavar='<policy>',
            help=_('Name or Id of the policy to show')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        senlin_client = self.app.client_manager.clustering
        return _show_policy(senlin_client, policy_id=parsed_args.policy)


def _show_policy(senlin_client, policy_id):
    try:
        policy = senlin_client.get_policy(policy_id)
    except sdk_exc.ResourceNotFound:
        raise exc.CommandError(_('Policy not found: %s') % policy_id)

    formatters = {
        'spec': senlin_utils.json_formatter
    }

    data = policy.to_dict()
    columns = sorted(data.keys())
    return columns, utils.get_dict_properties(data, columns,
                                              formatters=formatters)


class CreatePolicy(command.ShowOne):
    """Create a policy."""

    log = logging.getLogger(__name__ + ".CreatePolicy")

    def get_parser(self, prog_name):
        parser = super(CreatePolicy, self).get_parser(prog_name)
        parser.add_argument(
            '--spec-file',
            metavar='<spec-file>',
            required=True,
            help=_('The spec file used to create the policy')
        )
        parser.add_argument(
            'name',
            metavar='<name>',
            help=_('Name of the policy to create')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        senlin_client = self.app.client_manager.clustering
        spec = senlin_utils.get_spec_content(parsed_args.spec_file)
        attrs = {
            'name': parsed_args.name,
            'spec': spec,
        }

        policy = senlin_client.create_policy(**attrs)
        return _show_policy(senlin_client, policy.id)


class UpdatePolicy(command.ShowOne):
    """Update a policy."""

    log = logging.getLogger(__name__ + ".UpdatePolicy")

    def get_parser(self, prog_name):
        parser = super(UpdatePolicy, self).get_parser(prog_name)
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('New name of the policy to be updated')
        )
        parser.add_argument(
            'policy',
            metavar='<policy>',
            help=_('Name or ID of the policy to be updated')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        senlin_client = self.app.client_manager.clustering
        params = {
            'name': parsed_args.name,
        }
        policy = senlin_client.find_policy(parsed_args.policy)
        if policy is None:
            raise exc.CommandError(_('Policy not found: %s') %
                                   parsed_args.policy)
        senlin_client.update_policy(policy.id, **params)
        return _show_policy(senlin_client, policy_id=policy.id)


class DeletePolicy(command.Command):
    """Delete policy(s)."""

    log = logging.getLogger(__name__ + ".DeletePolicy")

    def get_parser(self, prog_name):
        parser = super(DeletePolicy, self).get_parser(prog_name)
        parser.add_argument(
            'policy',
            metavar='<policy>',
            nargs='+',
            help=_('Name or ID of policy(s) to delete')
        )
        parser.add_argument(
            '--force',
            action='store_true',
            help=_('Skip yes/no prompt (assume yes)')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering
        try:
            if not parsed_args.force and sys.stdin.isatty():
                sys.stdout.write(
                    _("Are you sure you want to delete this policy(s)"
                      " [y/N]?"))
                sys.stdout.flush()
                prompt_response = sys.stdin.readline().lower()
                if not prompt_response.startswith('y'):
                    return
        except KeyboardInterrupt:  # Ctrl-c
            self.log.info('Ctrl-c detected.')
            return
        except EOFError:  # Ctrl-d
            self.log.info('Ctrl-d detected')
            return

        failure_count = 0

        for pid in parsed_args.policy:
            try:
                senlin_client.delete_policy(pid, False)
            except Exception as ex:
                failure_count += 1
                print(ex)
        if failure_count:
            raise exc.CommandError(_('Failed to delete %(count)s of the '
                                     '%(total)s specified policy(s).') %
                                   {'count': failure_count,
                                   'total': len(parsed_args.policy)})
        print('Policy deleted: %s' % parsed_args.policy)


class ValidatePolicy(command.ShowOne):
    """Validate a policy."""

    log = logging.getLogger(__name__ + ".ValidatePolicy")

    def get_parser(self, prog_name):
        parser = super(ValidatePolicy, self).get_parser(prog_name)
        parser.add_argument(
            '--spec-file',
            metavar='<spec-file>',
            required=True,
            help=_('The spec file of the policy to be validated')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        senlin_client = self.app.client_manager.clustering
        spec = senlin_utils.get_spec_content(parsed_args.spec_file)
        attrs = {
            'spec': spec,
        }

        policy = senlin_client.validate_policy(**attrs)
        formatters = {
            'spec': senlin_utils.json_formatter
        }
        columns = [
            'created_at',
            'data',
            'domain',
            'id',
            'name',
            'project',
            'spec',
            'type',
            'updated_at',
            'user'
        ]
        return columns, utils.get_dict_properties(policy.to_dict(), columns,
                                                  formatters=formatters)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import yaml

from oslo_serialization import jsonutils

from osc_lib.tests import utils
from senlinclient.common import format_utils


columns = ['col1', 'col2', 'col3']
data = ['abcde', ['fg', 'hi', 'jk'], {'lmnop': 'qrstu'}]


class ShowJson(format_utils.JsonFormat):
    def take_action(self, parsed_args):
        return columns, data


class ShowYaml(format_utils.YamlFormat):
    def take_action(self, parsed_args):
        return columns, data


class ShowShell(format_utils.ShellFormat):
    def take_action(self, parsed_args):
        return columns, data


class ShowValue(format_utils.ValueFormat):
    def take_action(self, parsed_args):
        return columns, data


class TestFormats(utils.TestCommand):

    def test_json_format(self):
        self.cmd = ShowJson(self.app, None)
        parsed_args = self.check_parser(self.cmd, [], [])
        expected = jsonutils.dumps(dict(zip(columns, data)), indent=2)

        self.cmd.run(parsed_args)

        self.assertEqual(jsonutils.loads(expected),
                         jsonutils.loads(self.app.stdout.make_string()))

    def test_yaml_format(self):
        self.cmd = ShowYaml(self.app, None)
        parsed_args = self.check_parser(self.cmd, [], [])
        expected = yaml.safe_dump(dict(zip(columns, data)),
                                  default_flow_style=False)

        self.cmd.run(parsed_args)

        self.assertEqual(expected, self.app.stdout.make_string())

    def test_shell_format(self):
        self.cmd = ShowShell(self.app, None)
        parsed_args = self.check_parser(self.cmd, [], [])
        expected = '''\
col1="abcde"
col2="['fg', 'hi', 'jk']"
col3="{'lmnop': 'qrstu'}"
'''

        self.cmd.run(parsed_args)

        self.assertEqual(expected, self.app.stdout.make_string())

    def test_value_format(self):
        self.cmd = ShowValue(self.app, None)
        parsed_args = self.check_parser(self.cmd, [], [])
        expected = '''\
abcde
['fg', 'hi', 'jk']
{'lmnop': 'qrstu'}
'''

        self.cmd.run(parsed_args)

        self.assertEqual(expected, self.app.stdout.make_string())



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from senlinclient.tests.functional import base


class ReceiverTest(base.OpenStackClientTestBase):
    """Test for receivers"""

    def test_receiver_list(self):
        result = self.openstack('cluster receiver list')
        receiver_list = self.parser.listing(result)
        self.assertTableStruct(receiver_list, ['id', 'name', 'type',
                                               'cluster_id', 'action',
                                               'created_at'])

    def test_receiver_create(self):
        name = self.name_generate()
        pf = self.profile_create(name)
        self.addCleanup(self.profile_delete, pf['id'])
        cluster = self.cluster_create(pf['id'], name)
        self.addCleanup(self.cluster_delete, cluster['id'])
        receiver = self.receiver_create(name, cluster['id'])
        self.addCleanup(self.receiver_delete, receiver['id'])
        self.assertEqual(receiver['name'], name)
        self.assertEqual(receiver['type'], 'webhook')
        self.assertEqual(receiver['action'], 'CLUSTER_SCALE_OUT')

    def test_receiver_update(self):
        old_name = self.name_generate()
        pf = self.profile_create(old_name)
        self.addCleanup(self.profile_delete, pf['id'])
        cluster = self.cluster_create(pf['id'], old_name)
        self.addCleanup(self.cluster_delete, cluster['id'])
        receiver = self.receiver_create(old_name, cluster['id'])
        self.addCleanup(self.receiver_delete, receiver['id'])
        new_name = self.name_generate()

        cmd = ('cluster receiver update --name %s --params count=2 '
               '--action CLUSTER_SCALE_IN %s' % (new_name, receiver['id']))
        self.openstack(cmd)
        receiver_raw = self.openstack('cluster receiver show %s'
                                      % receiver['id'])
        receiver_data = self.show_to_dict(receiver_raw)
        self.assertNotEqual(receiver['name'], receiver_data['name'])
        self.assertEqual(receiver_data['name'], new_name)
        self.assertNotEqual(receiver['action'], receiver_data['action'])
        self.assertEqual(receiver_data['action'], 'CLUSTER_SCALE_IN')



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from senlinclient.tests.functional import base


class ClusterPolicyTest(base.OpenStackClientTestBase):
    """Test cluster for policy"""

    def test_cluster_policy_attach_and_detach(self):
        name = self.name_generate()
        po = self.policy_create(name)
        self.addCleanup(self.policy_delete, po['id'])
        pf = self.profile_create(name)
        self.addCleanup(self.profile_delete, pf['id'])
        cluster = self.cluster_create(pf['id'], name)
        self.addCleanup(self.cluster_delete, cluster['id'])

        cp_raw = self.openstack('cluster policy binding list %s'
                                % cluster['id'])
        cp_data = self.show_to_dict(cp_raw)
        self.assertEqual({}, cp_data)

        # Attach policy to cluster
        cmd = ('cluster policy attach --policy %s %s' % (po['id'],
                                                         cluster['id']))
        self.openstack(cmd)
        self.wait_for_status(cluster['id'], 'ACTIVE', 'cluster', 120)
        cmd = ('cluster policy binding show --policy %s %s' %
               (po['id'], cluster['id']))
        policy_raw = self.openstack(cmd)
        policy_data = self.show_to_dict(policy_raw)
        self.assertEqual(po['name'], policy_data['policy_name'])
        self.assertEqual(cluster['name'], policy_data['cluster_name'])
        self.assertTrue(policy_data['is_enabled'])

        # Detach policy from cluster
        cmd = ('cluster policy detach --policy %s %s' % (po['id'],
                                                         cluster['id']))
        self.openstack(cmd)
        self.wait_for_status(cluster['id'], 'ACTIVE', 'cluster', 120)
        cp_raw = self.openstack('cluster policy binding list %s'
                                % cluster['id'])
        cp_data = self.show_to_dict(cp_raw)
        self.assertEqual({}, cp_data)

    def test_cluster_policy_list(self):
        name = self.name_generate()
        po = self.policy_create(name)
        self.addCleanup(self.policy_delete, po['id'])
        pf = self.profile_create(name)
        self.addCleanup(self.profile_delete, pf['id'])
        cluster = self.cluster_create(pf['id'], name)
        self.addCleanup(self.cluster_delete, cluster['id'])

        cmd = ('cluster policy attach --policy %s %s' % (po['id'],
                                                         cluster['id']))
        self.openstack(cmd)
        self.wait_for_status(cluster['id'], 'ACTIVE', 'cluster', 120)

        # List cluster policy binding
        cmd = ('cluster policy binding list --filters policy_name=%s %s'
               % (po['name'], cluster['id']))
        result = self.openstack(cmd)
        binding_list = self.parser.listing(result)
        self.assertTableStruct(binding_list, ['policy_id', 'policy_name',
                                              'policy_type', 'is_enabled'])
        cmd = ('cluster policy detach --policy %s %s' % (po['id'],
                                                         cluster['id']))
        self.openstack(cmd)
        self.wait_for_status(cluster['id'], 'ACTIVE', 'cluster', 120)

    def test_cluster_policy_update(self):
        name = self.name_generate()
        po = self.policy_create(name)
        self.addCleanup(self.policy_delete, po['id'])
        pf = self.profile_create(name)
        self.addCleanup(self.profile_delete, pf['id'])
        cluster = self.cluster_create(pf['id'], name)
        self.addCleanup(self.cluster_delete, cluster['id'])

        cmd = ('cluster policy attach --policy %s %s' % (po['id'],
                                                         cluster['id']))
        self.openstack(cmd)
        self.wait_for_status(cluster['id'], 'ACTIVE', 'cluster', 120)

        # Update cluster policy binding
        cmd = ('cluster policy binding update --policy %s --enabled false %s'
               % (po['id'], cluster['id']))
        self.openstack(cmd)
        self.wait_for_status(cluster['id'], 'ACTIVE', 'cluster', 120)

        cp_update = self.openstack('cluster policy binding show --policy %s %s'
                                   % (po['id'], cluster['id']))
        cp_update_data = self.show_to_dict(cp_update)
        self.assertFalse(cp_update_data['is_enabled'].isupper())
        cmd = ('cluster policy detach --policy %s %s' % (po['id'],
                                                         cluster['id']))
        self.openstack(cmd)
        self.wait_for_status(cluster['id'], 'ACTIVE', 'cluster', 120)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Clustering v1 action implementations"""

import logging

from openstack import exceptions as sdk_exc
from osc_lib.command import command
from osc_lib import exceptions as exc
from osc_lib import utils

from senlinclient.common.i18n import _
from senlinclient.common import utils as senlin_utils


class ListAction(command.Lister):
    """List actions."""

    log = logging.getLogger(__name__ + ".ListAction")

    def get_parser(self, prog_name):
        parser = super(ListAction, self).get_parser(prog_name)
        parser.add_argument(
            '--filters',
            metavar='<"key1=value1;key2=value2...">',
            help=_("Filter parameters to apply on returned actions. "
                   "This can be specified multiple times, or once with "
                   "parameters separated by a semicolon. The valid filter "
                   "keys are: ['name', 'target', 'action', 'status']. "
                   "NOTICE: The value of 'target', if provided, "
                   "must be a full ID."),
            action='append'
        )
        parser.add_argument(
            '--sort',
            metavar='<key>[:<direction>]',
            help=_("Sorting option which is a string containing a list of "
                   "keys separated by commas. Each key can be optionally "
                   "appended by a sort direction (:asc or :desc). The valid "
                   "sort keys are: ['name', 'target', 'action', 'created_at',"
                   " 'status']")
        )
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            help=_('Limit the number of actions returned')
        )
        parser.add_argument(
            '--marker',
            metavar='<id>',
            help=_('Only return actions that appear after the given action ID')
        )
        parser.add_argument(
            '--global-project',
            default=False,
            action="store_true",
            help=_('Whether actions from all projects should be listed. '
                   ' Default to False. Setting this to True may demand '
                   'for an admin privilege')
        )
        parser.add_argument(
            '--full-id',
            default=False,
            action="store_true",
            help=_('Print full IDs in list')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        senlin_client = self.app.client_manager.clustering

        columns = ['id', 'name', 'action', 'status', 'target_id', 'depends_on',
                   'depended_by', 'created_at', 'cluster_id']

        queries = {
            'sort': parsed_args.sort,
            'limit': parsed_args.limit,
            'marker': parsed_args.marker,
            'global_project': parsed_args.global_project,
        }

        if parsed_args.filters:
            queries.update(senlin_utils.format_parameters(parsed_args.filters))

        actions = senlin_client.actions(**queries)

        formatters = {}
        s = None
        if not parsed_args.full_id:
            s = 8
            formatters['id'] = lambda x: x[:s]
            formatters['target_id'] = lambda x: x[:s]

        formatters['depends_on'] = lambda x: '\n'.join(a[:s] for a in x)
        formatters['depended_by'] = lambda x: '\n'.join(a[:s] for a in x)

        return (
            columns,
            (utils.get_item_properties(a, columns,
                                       formatters=formatters)
             for a in actions)
        )


def _show_action(senlin_client, action_id):
    try:
        action = senlin_client.get_action(action_id)
    except sdk_exc.ResourceNotFound:
        raise exc.CommandError(_('Action not found: %s')
                               % action_id)

    formatters = {
        'inputs': senlin_utils.json_formatter,
        'outputs': senlin_utils.json_formatter,
        'metadata': senlin_utils.json_formatter,
        'data': senlin_utils.json_formatter,
        'depends_on': senlin_utils.list_formatter,
        'depended_by': senlin_utils.list_formatter,
    }
    data = action.to_dict()
    columns = sorted(data.keys())
    return columns, utils.get_dict_properties(data, columns,
                                              formatters=formatters)


class ShowAction(command.ShowOne):
    """Show detailed info about the specified action."""

    log = logging.getLogger(__name__ + ".ShowAction")

    def get_parser(self, prog_name):
        parser = super(ShowAction, self).get_parser(prog_name)
        parser.add_argument(
            'action',
            metavar='<action>',
            help=_('Name or ID of the action to show the details for')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)

        senlin_client = self.app.client_manager.clustering
        return _show_action(senlin_client, parsed_args.action)


class UpdateAction(command.ShowOne):
    """Update an action."""

    log = logging.getLogger(__name__ + ".UpdateAction")

    def get_parser(self, prog_name):
        parser = super(UpdateAction, self).get_parser(prog_name)
        parser.add_argument(
            '--status',
            metavar='<status>',
            help=_('The new status for the action')
        )
        parser.add_argument(
            'action',
            metavar='<action>',
            help=_('ID of the action to update')
        )
        return parser

    def take_action(self, parsed_args):
        self.log.debug("take_action(%s)", parsed_args)
        senlin_client = self.app.client_manager.clustering

        params = {
            'status': parsed_args.status,
        }

        senlin_client.update_action(parsed_args.action, **params)
        return _show_action(senlin_client, parsed_args.action)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
from unittest import mock

from openstack import exceptions as sdk_exc
from osc_lib import exceptions as exc

from senlinclient.tests.unit.v1 import fakes
from senlinclient.v1 import action as osc_action


class TestAction(fakes.TestClusteringv1):

    def setUp(self):
        super(TestAction, self).setUp()
        self.mock_client = self.app.client_manager.clustering


class TestActionList(TestAction):

    columns = ['id', 'name', 'action', 'status', 'target_id', 'depends_on',
               'depended_by', 'created_at', 'cluster_id']
    defaults = {
        'global_project': False,
        'marker': None,
        'limit': None,
        'sort': None,
    }

    def setUp(self):
        super(TestActionList, self).setUp()
        self.cmd = osc_action.ListAction(self.app, None)
        fake_action = mock.Mock(
            action="NODE_CREATE",
            cluster_id="FAKE_CLUSTER_ID",
            cause="RPC Request",
            created_at="2015-12-04T04:54:41",
            depended_by=[],
            depends_on=[],
            end_time=1425550000.0,
            id="2366d440-c73e-4961-9254-6d1c3af7c167",
            inputs={},
            interval=-1,
            name="node_create_0df0931b",
            outputs={},
            owner=None,
            start_time=1425550000.0,
            status="SUCCEEDED",
            status_reason="Action completed successfully.",
            target_id="0df0931b-e251-4f2e-8719-4ebfda3627ba",
            timeout=3600,
            updated_at=None
        )
        fake_action.to_dict = mock.Mock(return_value={})
        self.mock_client.actions = mock.Mock(return_value=[fake_action])

    def test_action_list_defaults(self):
        arglist = []
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.actions.assert_called_with(**self.defaults)
        self.assertEqual(self.columns, columns)

    def test_action_list_full_id(self):
        arglist = ['--full-id']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.actions.assert_called_with(**self.defaults)
        self.assertEqual(self.columns, columns)

    def test_action_list_limit(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['limit'] = '3'
        arglist = ['--limit', '3']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.actions.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)

    def test_action_list_sort(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['sort'] = 'name:asc'
        arglist = ['--sort', 'name:asc']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.actions.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)

    def test_action_list_sort_invalid_key(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['sort'] = 'bad_key'
        arglist = ['--sort', 'bad_key']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.mock_client.actions.side_effect = sdk_exc.HttpException()
        self.assertRaises(sdk_exc.HttpException,
                          self.cmd.take_action, parsed_args)

    def test_action_list_sort_invalid_direction(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['sort'] = 'name:bad_direction'
        arglist = ['--sort', 'name:bad_direction']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.mock_client.actions.side_effect = sdk_exc.HttpException()
        self.assertRaises(sdk_exc.HttpException,
                          self.cmd.take_action, parsed_args)

    def test_action_list_filter(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['name'] = 'my_action'
        arglist = ['--filter', 'name=my_action']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.actions.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)

    def test_action_list_marker(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['marker'] = 'a9448bf6'
        arglist = ['--marker', 'a9448bf6']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.actions.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)


class TestActionShow(TestAction):

    def setUp(self):
        super(TestActionShow, self).setUp()
        self.cmd = osc_action.ShowAction(self.app, None)
        fake_action = mock.Mock(
            action="NODE_CREATE",
            cluster_id="FAKE_CLUSTER_ID",
            cause="RPC Request",
            created_at="2015-12-04T04:54:41",
            depended_by=[],
            depends_on=[],
            end_time=1425550000.0,
            id="2366d440-c73e-4961-9254-6d1c3af7c167",
            inputs={},
            interval=-1,
            name="node_create_0df0931b",
            outputs={},
            owner=None,
            start_time=1425550000.0,
            status="SUCCEEDED",
            status_reason="Action completed successfully.",
            target_id="0df0931b-e251-4f2e-8719-4ebfda3627ba",
            timeout=3600,
            updated_at=None
        )
        fake_action.to_dict = mock.Mock(return_value={})
        self.mock_client.get_action = mock.Mock(return_value=fake_action)

    def test_action_show(self):
        arglist = ['my_action']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.get_action.assert_called_with('my_action')

    def test_action_show_not_found(self):
        arglist = ['my_action']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.mock_client.get_action.side_effect = sdk_exc.ResourceNotFound()
        error = self.assertRaises(exc.CommandError, self.cmd.take_action,
                                  parsed_args)
        self.assertEqual('Action not found: my_action', str(error))


class TestActionUpdate(TestAction):

    def setUp(self):
        super(TestActionUpdate, self).setUp()
        self.cmd = osc_action.UpdateAction(self.app, None)
        fake_action = mock.Mock(
            action="NODE_CREATE",
            cluster_id="FAKE_CLUSTER_ID",
            cause="RPC Request",
            created_at="2015-12-04T04:54:41",
            depended_by=[],
            depends_on=[],
            end_time=1425550000.0,
            id="2366d440-c73e-4961-9254-6d1c3af7c167",
            inputs={},
            interval=-1,
            name="node_create_0df0931b",
            outputs={},
            owner=None,
            start_time=1425550000.0,
            status="INIT",
            status_reason="Action completed successfully.",
            target_id="0df0931b-e251-4f2e-8719-4ebfda3627ba",
            timeout=3600,
            updated_at=None
        )
        fake_action.to_dict = mock.Mock(return_value={})
        self.mock_client.get_action = mock.Mock(return_value=fake_action)
        self.mock_client.update_action = mock.Mock(return_value=fake_action)

    def test_action_update(self):
        arglist = ['--status', 'CANCELLED',
                   '2366d440-c73e-4961-9254-6d1c3af7c167']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        defaults = {
            "status": "CANCELLED"
        }

        self.cmd.take_action(parsed_args)

        self.mock_client.update_action.assert_called_with(
            "2366d440-c73e-4961-9254-6d1c3af7c167", **defaults)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy
import io
from unittest import mock

from openstack import exceptions as sdk_exc
from osc_lib import exceptions as exc

from senlinclient.tests.unit.v1 import fakes
from senlinclient.v1 import policy as osc_policy


class TestPolicy(fakes.TestClusteringv1):
    def setUp(self):
        super(TestPolicy, self).setUp()
        self.mock_client = self.app.client_manager.clustering


class TestPolicyList(TestPolicy):
    columns = ['id', 'name', 'type', 'created_at']
    response = {"policies": [
        {
        }
    ]}
    defaults = {
        'global_project': False,
        'marker': None,
        'limit': None,
        'sort': None,
    }

    def setUp(self):
        super(TestPolicyList, self).setUp()
        self.cmd = osc_policy.ListPolicy(self.app, None)
        fake_policy = mock.Mock(
            created_at="2015-02-15T08:33:13.000000",
            data={},
            domain=None,
            id="7192d8df-73be-4e98-ab99-1cf6d5066729",
            project_id="42d9e9663331431f97b75e25136307ff",
            spec={
                "description": "A test policy",
                "properties": {
                    "criteria": "OLDEST_FIRST",
                    "destroy_after_deletion": True,
                    "grace_period": 60,
                    "reduce_desired_capacity": False
                },
                "type": "senlin.policy.deletion",
                "version": "1.0"
            },
            type="senlin.policy.deletion-1.0",
            updated_at=None,
            user_id="5e5bf8027826429c96af157f68dc9072"
        )
        fake_policy.name = "test_policy_1"
        fake_policy.to_dict = mock.Mock(return_value={})
        self.mock_client.policies = mock.Mock(
            return_value=self.response)

    def test_policy_list_defaults(self):
        arglist = []
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.policies.assert_called_with(**self.defaults)
        self.assertEqual(self.columns, columns)

    def test_policy_list_full_id(self):
        arglist = ['--full-id']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.policies.assert_called_with(**self.defaults)
        self.assertEqual(self.columns, columns)

    def test_policy_list_limit(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['limit'] = '3'
        arglist = ['--limit', '3']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.policies.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)

    def test_policy_list_sort(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['sort'] = 'name:asc'
        arglist = ['--sort', 'name:asc']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.policies.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)

    def test_policy_list_sort_invalid_key(self):
        self.mock_client.policies = mock.Mock(
            return_value=self.response)
        kwargs = copy.deepcopy(self.defaults)
        kwargs['sort'] = 'bad_key'
        arglist = ['--sort', 'bad_key']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.mock_client.policies.side_effect = sdk_exc.HttpException()
        self.assertRaises(sdk_exc.HttpException,
                          self.cmd.take_action, parsed_args)

    def test_policy_list_sort_invalid_direction(self):
        self.mock_client.policies = mock.Mock(
            return_value=self.response)
        kwargs = copy.deepcopy(self.defaults)
        kwargs['sort'] = 'name:bad_direction'
        arglist = ['--sort', 'name:bad_direction']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.mock_client.policies.side_effect = sdk_exc.HttpException()
        self.assertRaises(sdk_exc.HttpException,
                          self.cmd.take_action, parsed_args)

    def test_policy_list_marker(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['marker'] = 'a9448bf6'
        arglist = ['--marker', 'a9448bf6']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.policies.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)

    def test_policy_list_filter(self):
        kwargs = copy.deepcopy(self.defaults)
        kwargs['name'] = 'my_policy'
        arglist = ['--filter', 'name=my_policy']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        columns, data = self.cmd.take_action(parsed_args)
        self.mock_client.policies.assert_called_with(**kwargs)
        self.assertEqual(self.columns, columns)


class TestPolicyShow(TestPolicy):
    response = {"policy": {
    }}

    def setUp(self):
        super(TestPolicyShow, self).setUp()
        self.cmd = osc_policy.ShowPolicy(self.app, None)
        fake_policy = mock.Mock(
            created_at="2015-03-02T07:40:31",
            data={},
            domain_id=None,
            id="02f62195-2198-4797-b0a9-877632208527",
            project_id="42d9e9663331431f97b75e25136307ff",
            spec={
                "properties": {
                    "adjustment": {
                        "best_effort": True,
                        "min_step": 1,
                        "number": 1,
                        "type": "CHANGE_IN_CAPACITY"
                    },
                    "event": "CLUSTER_SCALE_IN"
                },
                "type": "senlin.policy.scaling",
                "version": "1.0"
            },
            type="senlin.policy.scaling-1.0",
            updated_at=None,
            user_id="5e5bf8027826429c96af157f68dc9072"
        )
        fake_policy.name = "sp001"
        fake_policy.to_dict = mock.Mock(return_value={})
        self.mock_client.get_policy = mock.Mock(return_value=fake_policy)

    def test_policy_show(self):
        arglist = ['sp001']
        parsed_args = self.check_parser(self.cmd, arglist, [])

        self.cmd.take_action(parsed_args)

        self.mock_client.get_policy.assert_called_with('sp001')
        policy = self.mock_client.get_policy('sp001')
        self.assertEqual("2015-03-02T07:40:31", policy.created_at)
        self.assertEqual({}, policy.data)
        self.assertEqual("02f62195-2198-4797-b0a9-877632208527", policy.id)
        self.assertEqual("sp001", policy.name)
        self.assertEqual("42d9e9663331431f97b75e25136307ff", policy.project_id)
        self.assertEqual("senlin.policy.scaling-1.0", policy.type)
        self.assertIsNone(policy.updated_at)

    def test_policy_show_not_found(self):
        arglist = ['sp001']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.mock_client.get_policy.side_effect = sdk_exc.ResourceNotFound()
        self.assertRaises(exc.CommandError, self.cmd.take_action, parsed_args)


class TestPolicyCreate(TestPolicy):
    spec_path = 'senlinclient/tests/test_specs/deletion_policy.yaml'
    defaults = {
        "name": "my_policy",
        "spec": {
            "version": 1,
            "type": "senlin.policy.deletion",
            "description": "A policy for choosing victim node(s) from a "
                           "cluster for deletion.",
            "properties": {
                "destroy_after_deletion": True,
                "grace_period": 60,
                "reduce_desired_capacity": False,
                "criteria": "OLDEST_FIRST"
            }
        }
    }

    def setUp(self):
        super(TestPolicyCreate, self).setUp()
        self.cmd = osc_policy.CreatePolicy(self.app, None)
        fake_policy = mock.Mock(
            created_at="2016-02-21T02:38:36",
            data={},
            domain_id=None,
            id="9f779ddf-744e-48bd-954c-acef7e11116c",
            project_id="5f1cc92b578e4e25a3b284179cf20a9b",
            spec={
                "description": "A policy for choosing victim node(s) from a "
                               "cluster for deletion.",
                "properties": {
                    "criteria": "OLDEST_FIRST",
                    "destroy_after_deletion": True,
                    "grace_period": 60,
                    "reduce_desired_capacity": False
                },
                "type": "senlin.policy.deletion",
                "version": 1.0
            },
            type="senlin.policy.deletion-1.0",
            updated_at=None,
            user_id="2d7aca950f3e465d8ef0c81720faf6ff"
        )
        fake_policy.name = "my_policy"
        fake_policy.to_dict = mock.Mock(return_value={})
        self.mock_client.create_policy = mock.Mock(return_value=fake_policy)
        self.mock_client.get_policy = mock.Mock(return_value=fake_policy)

    def test_policy_create_defaults(self):
        arglist = ['my_policy', '--spec-file', self.spec_path]
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.create_policy.assert_called_with(**self.defaults)


class TestPolicyUpdate(TestPolicy):

    def setUp(self):
        super(TestPolicyUpdate, self).setUp()
        self.cmd = osc_policy.UpdatePolicy(self.app, None)
        fake_policy = mock.Mock(
            created_at="2016-02-21T02:38:36",
            data={},
            domain_id=None,
            id="9f779ddf-744e-48bd-954c-acef7e11116c",
            project_id="5f1cc92b578e4e25a3b284179cf20a9b",
            spec={
                "description": "A policy for choosing victim node(s) from a "
                               "cluster for deletion.",
                "properties": {
                    "criteria": "OLDEST_FIRST",
                    "destroy_after_deletion": True,
                    "grace_period": 60,
                    "reduce_desired_capacity": False
                },
                "type": "senlin.policy.deletion",
                "version": 1.0
            },
            type="senlin.policy.deletion-1.0",
            updated_at=None,
            user_id="2d7aca950f3e465d8ef0c81720faf6ff"
        )
        fake_policy.name = "new_policy"
        fake_policy.to_dict = mock.Mock(return_value={})
        self.mock_client.update_policy = mock.Mock(return_value=fake_policy)
        self.mock_client.get_policy = mock.Mock(return_value=fake_policy)
        self.mock_client.find_policy = mock.Mock(return_value=fake_policy)

    def test_policy_update_defaults(self):
        arglist = ['--name', 'new_policy', '9f779ddf']
        parsed_args = self.check_parser(self.cmd, arglist, [])

        self.cmd.take_action(parsed_args)

        self.mock_client.update_policy.assert_called_with(
            '9f779ddf-744e-48bd-954c-acef7e11116c', name="new_policy")

    def test_policy_update_not_found(self):
        arglist = ['--name', 'new_policy', 'c6b8b252']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.mock_client.find_policy.return_value = None
        error = self.assertRaises(exc.CommandError,
                                  self.cmd.take_action,
                                  parsed_args)
        self.assertIn('Policy not found: c6b8b252', str(error))


class TestPolicyDelete(TestPolicy):
    def setUp(self):
        super(TestPolicyDelete, self).setUp()
        self.cmd = osc_policy.DeletePolicy(self.app, None)
        self.mock_client.delete_policy = mock.Mock()

    def test_policy_delete(self):
        arglist = ['policy1', 'policy2', 'policy3']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.delete_policy.assert_has_calls(
            [mock.call('policy1', False), mock.call('policy2', False),
             mock.call('policy3', False)]
        )

    def test_policy_delete_force(self):
        arglist = ['policy1', 'policy2', 'policy3', '--force']
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.delete_policy.assert_has_calls(
            [mock.call('policy1', False), mock.call('policy2', False),
             mock.call('policy3', False)]
        )

    def test_policy_delete_not_found(self):
        arglist = ['my_policy']
        self.mock_client.delete_policy.side_effect = sdk_exc.ResourceNotFound
        parsed_args = self.check_parser(self.cmd, arglist, [])
        error = self.assertRaises(exc.CommandError, self.cmd.take_action,
                                  parsed_args)
        self.assertIn('Failed to delete 1 of the 1 specified policy(s).',
                      str(error))

    def test_policy_delete_one_found_one_not_found(self):
        arglist = ['policy1', 'policy2']
        self.mock_client.delete_policy.side_effect = (
            [None, sdk_exc.ResourceNotFound]
        )
        parsed_args = self.check_parser(self.cmd, arglist, [])
        error = self.assertRaises(exc.CommandError,
                                  self.cmd.take_action, parsed_args)
        self.mock_client.delete_policy.assert_has_calls(
            [mock.call('policy1', False), mock.call('policy2', False)]
        )
        self.assertEqual('Failed to delete 1 of the 2 specified policy(s).',
                         str(error))

    @mock.patch('sys.stdin', spec=io.StringIO)
    def test_policy_delete_prompt_yes(self, mock_stdin):
        arglist = ['my_policy']
        mock_stdin.isatty.return_value = True
        mock_stdin.readline.return_value = 'y'
        parsed_args = self.check_parser(self.cmd, arglist, [])

        self.cmd.take_action(parsed_args)

        mock_stdin.readline.assert_called_with()
        self.mock_client.delete_policy.assert_called_with('my_policy',
                                                          False)

    @mock.patch('sys.stdin', spec=io.StringIO)
    def test_policy_delete_prompt_no(self, mock_stdin):
        arglist = ['my_policy']
        mock_stdin.isatty.return_value = True
        mock_stdin.readline.return_value = 'n'
        parsed_args = self.check_parser(self.cmd, arglist, [])

        self.cmd.take_action(parsed_args)

        mock_stdin.readline.assert_called_with()
        self.mock_client.delete_policy.assert_not_called()


class TestPolicyValidate(TestPolicy):
    spec_path = 'senlinclient/tests/test_specs/deletion_policy.yaml'
    defaults = {
        "spec": {
            "version": 1,
            "type": "senlin.policy.deletion",
            "description": "A policy for choosing victim node(s) from a "
                           "cluster for deletion.",
            "properties": {
                "destroy_after_deletion": True,
                "grace_period": 60,
                "reduce_desired_capacity": False,
                "criteria": "OLDEST_FIRST"
            }
        }
    }

    def setUp(self):
        super(TestPolicyValidate, self).setUp()
        self.cmd = osc_policy.ValidatePolicy(self.app, None)
        fake_policy = mock.Mock(
            created_at=None,
            data={},
            domain_id=None,
            id=None,
            project_id="5f1cc92b578e4e25a3b284179cf20a9b",
            spec={
                "description": "A policy for choosing victim node(s) from a "
                               "cluster for deletion.",
                "properties": {
                    "criteria": "OLDEST_FIRST",
                    "destroy_after_deletion": True,
                    "grace_period": 60,
                    "reduce_desired_capacity": False
                },
                "type": "senlin.policy.deletion",
                "version": 1.0
            },
            type="senlin.policy.deletion-1.0",
            updated_at=None,
            user_id="2d7aca950f3e465d8ef0c81720faf6ff"
        )
        fake_policy.name = "validated_policy"
        fake_policy.to_dict = mock.Mock(return_value={})
        self.mock_client.validate_policy = mock.Mock(return_value=fake_policy)

    def test_policy_validate(self):
        arglist = ['--spec-file', self.spec_path]
        parsed_args = self.check_parser(self.cmd, arglist, [])
        self.cmd.take_action(parsed_args)
        self.mock_client.validate_policy.assert_called_with(**self.defaults)

        policy = self.mock_client.validate_policy(**self.defaults)

        self.assertEqual("5f1cc92b578e4e25a3b284179cf20a9b", policy.project_id)
        self.assertEqual({}, policy.data)
        self.assertIsNone(policy.id)
        self.assertEqual("validated_policy", policy.name)
        self.assertEqual("senlin.policy.deletion-1.0", policy.type)
        self.assertIsNone(policy.updated_at)




# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
# python-senlinclient documentation build configuration file

# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'openstackdocstheme',
]

# The content that will be inserted into the main body of an autoclass
# directive.
autoclass_content = 'both'

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# openstackdocstheme options
openstackdocs_repo_name = 'openstack/python-senlinclient'
openstackdocs_bug_project = 'python-senlinclient'
openstackdocs_bug_tag = ''
copyright = 'OpenStack Contributors'

# List of directories, relative to source directory, that shouldn't be searched
# for source files.
exclude_trees = []

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'native'

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages.  Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
html_theme = 'openstackdocs'

# -- Options for manual page output ------------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
    ('index', 'senlin', 'OpenStack Senlin command line client',
     ['OpenStack Contributors'], 1),
]
